import os
import calliope
import pytest  # pylint: disable=unused-import
import tempfile

from calliope.core.attrdict import AttrDict
from calliope.test.common.util import check_error_or_warning


HTML_STRINGS = AttrDict.from_yaml(
    os.path.join(os.path.dirname(__file__), 'common', 'html_strings.yaml')
)


class TestPlotting:
    @pytest.fixture(scope="module")
    def national_scale_example(self):
        model = calliope.examples.national_scale(
            override_dict={'model.subset_time': '2005-01-01'}
        )
        model.run()
        return model

    def test_national_scale_plotting(self, national_scale_example):
        model = national_scale_example

        plot_html_outputs = {
            'capacity': model.plot.capacity(html_only=True),
            'timeseries': model.plot.timeseries(html_only=True),
            'transmission': model.plot.transmission(html_only=True),
        }

        for plot_type in HTML_STRINGS['national_scale']:
            for string in HTML_STRINGS['national_scale'][plot_type]:
                assert string in plot_html_outputs[plot_type]

        # Also just try plotting the summary
        model.plot.summary()

        # Testing that the model can handle not having supply_plus technologies
        # Wrapped in temporary directory as we can't stop it saving an HTML file
        model._model_data = model._model_data.drop('resource_con')
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'test_plot.html')
            model.plot.timeseries(
                plotly_kwarg_updates={'auto_open': False, 'filename': out_path}
            )

    def test_milp_plotting(self):
        override = {'model.subset_time': '2005-07-01'}
        model = calliope.examples.milp(override_dict=override)
        model.run()

        plot_html_outputs = {
            'capacity': model.plot.capacity(html_only=True),
            'timeseries': model.plot.timeseries(html_only=True),
            'transmission': model.plot.transmission(html_only=True),
        }

        for plot_type in HTML_STRINGS['milp']:
            for string in HTML_STRINGS['milp'][plot_type]:
                assert string in plot_html_outputs[plot_type]

        # Also just try plotting the summary
        model.plot.summary()

    def test_failed_cap_plotting(self, national_scale_example):
        model = national_scale_example

        # should fail, not in array
        with pytest.raises(ValueError):
            model.plot.capacity(array='carrier_prod')
            model.plot.capacity(array=['energy_eff', 'energy_cap'])
            # orient has to be 'v', 'vertical', 'h', or 'horizontal'
            model.plot.capacity(orient='g')

    def test_failed_timeseries_plotting(self, national_scale_example):
        model = national_scale_example

        # should fail, not in array
        with pytest.raises(ValueError):
            model.plot.timeseries(array='energy_cap')
            model.plot.timeseries(squeeze=False)
            model.plot.timeseries(sum_dims=None)

    def test_clustered_plotting(self):
        override = {'model.time.function_options.k': 2}
        model = calliope.examples.time_clustering(override_dict=override)

        # While we have a model that hasn't been run, try plotting transmission and capacity
        model.plot.transmission(html_only=True)
        model.plot.capacity(html_only=True)

        model.run()
        plot_html = model.plot.timeseries(html_only=True)
        for string in HTML_STRINGS['clustering']['timeseries']:
            assert string in plot_html

    def test_subset_plotting(self, national_scale_example):
        model = national_scale_example

        model.plot.capacity(
            html_only=True, subset={'timesteps': ['2015-01-01 01:00']}
        )

        # should raise, subsetting with a tech that does not exist
        with pytest.raises(ValueError) as excinfo:
            model.plot.capacity(
                html_only=True, subset={'techs': ['foobar']}
            )

        assert check_error_or_warning(excinfo, 'No data to plot')

    def test_subset_array(self, national_scale_example):
        model = national_scale_example

        model.plot.capacity(html_only=True, array='inputs')
        model.plot.capacity(html_only=True, array='results')
        model.plot.capacity(html_only=True, array='energy_cap')
        model.plot.capacity(html_only=True, array='storage_cap')
        model.plot.capacity(
            html_only=True, array=['systemwide_levelised_cost', 'storage_cap']
        )

        model.plot.timeseries(html_only=True, array='inputs')
        model.plot.timeseries(html_only=True, array='results')
        model.plot.timeseries(html_only=True, array='power')
        model.plot.timeseries(html_only=True, array='resource')
        model.plot.timeseries(
            html_only=True, array=['resource_con', 'cost_var']
        )

    def test_long_name(self, national_scale_example):
            model = national_scale_example
            model._model_data['names'] = model._model_data.names.astype('<U100')
            model._model_data.names.loc['ccgt'] = (
                'a long name for a technology, longer than 30 characters'
            )
            model._model_data.names.loc['csp'] = (
                'a really very long name for a technology that is longer than 60 characters'
            )
            model._model_data.names.loc['battery'] = (
                'another_long_name_but_without_space_to_break_at'
            )
            model._model_data.names.loc['ac_transmission'] = (
                'long_transmission_name_which_has two break types in technology name'
            )

            broken_names = [
                'a long name for a technology,<br>longer than 30 characters',
                'another_long_name_but_without_...<br>space_to_break_at',
                'a really very long name for a<br>technology that is longer<br>than 60 characters'
            ]

            html_cap = model.plot.capacity(html_only=True)
            html_timeseries = model.plot.timeseries(html_only=True)
            html_transmission = model.plot.transmission(html_only=True)
            for i in broken_names:
                assert i in html_cap
                assert i in html_timeseries
            assert (
                'long_transmission_name_which_h...<br>as two break types in<br>technology name'
                in html_transmission
            )

    def test_plot_cost(self):
        model = calliope.examples.national_scale(
            override_dict={
                'techs.ccgt.costs.carbon': {'energy_cap': 10, 'interest_rate': 0.01}
            }
        )

        model.run()

        # should fail, multiple costs provided, can only plot one
        with pytest.raises(ValueError):
            model.plot.capacity(html_only=True, array='results')

        # should succeed, multiple costs provided, subset to one
        model.plot.capacity(
            html_only=True, array='results', subset={'costs': 'carbon'}
        )

        # FIXME: sum_dims doesn't seem to work at all
        # model.plot.capacity(html_only=True, sum_dims=['locs'])

    def test_to_file(self, national_scale_example):
        model = national_scale_example

        # should fail, 'gif' not in allowed extensions
        with pytest.raises(TypeError):
            model.plot.capacity(
                to_file='plot_to_save.gif',
                plotly_kwarg_updates={'auto_open': False})

        # FIXME: currently throws up save dialogue rather than just
        # saving the file

        #with tempfile.TemporaryDirectory() as tempdir:
        #    for extension in ['png', 'jpeg', 'svg', 'webp']:
        #        out_path = os.path.join(tempdir, 'plot_to_save.' + extension)
        #        model.plot.capacity(array='energy_cap', to_file=out_path, auto_open=False)
        #        assert os.path.isfile(out_path)

        # should fail, cannot save a plot with multiple DataArrays being plotted
        with pytest.raises(ValueError):
            model.plot.capacity(
                to_file='plot_to_save.svg',
                plotly_kwarg_updates={'auto_open': False})

        # test saving summary to file
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'test_summary.html')
            model.plot.summary(to_file=out_path)
            assert os.path.isfile(out_path)



import pytest
from pytest import approx

import calliope

_OVERRIDE_FILE = calliope.examples._PATHS['national_scale'] + '/overrides.yaml'


class TestNationalScaleExampleModelSenseChecks:
    def test_group_share_prod_min(self):
        model = calliope.examples.national_scale(
            override_file=_OVERRIDE_FILE + ':cold_fusion,group_share_cold_fusion_prod'
        )
        model.run()

        df_carrier_prod = (
            model.get_formatted_array('carrier_prod')
                 .loc[dict(carriers='power')].sum('locs').sum('timesteps')
                 .to_pandas()
        )

        prod_share = (
            df_carrier_prod.loc[['cold_fusion', 'csp']].sum() /
            df_carrier_prod.loc[['ccgt', 'cold_fusion', 'csp']].sum()
        )

        assert prod_share == approx(0.85)

    def test_group_share_cap_max(self):
        model = calliope.examples.national_scale(
            override_file=_OVERRIDE_FILE + ':cold_fusion,group_share_cold_fusion_cap'
        )
        model.run()

        cap_share = (
            model.get_formatted_array('energy_cap').to_pandas().loc[:, ['cold_fusion', 'csp']].sum().sum() /
            model.get_formatted_array('energy_cap').to_pandas().loc[:, ['ccgt', 'cold_fusion', 'csp']].sum().sum()
        )

        assert cap_share == approx(0.2)

    def test_systemwide_equals(self):
        model = calliope.examples.national_scale(
            override_dict={'techs.ccgt.constraints.energy_cap_max_systemwide': 10000}
        )
        model.run()
        # Check that setting `_equals` to a finite value leads to forcing
        assert (
            model.get_formatted_array('energy_cap').loc[{'techs': 'ccgt'}].sum() == 10000
        )

    def test_reserve_margin(self):
        model = calliope.examples.national_scale(
            override_file=_OVERRIDE_FILE + ':reserve_margin'
        )

        model.run()

        # constraint_string = '-Inf : -1.1 * ( carrier_con[region1::demand_power::power,2005-01-05 16:00:00] + carrier_con[region2::demand_power::power,2005-01-05 16:00:00] ) / timestep_resolution[2005-01-05 16:00:00] - energy_cap[region1::ccgt] - energy_cap[region1-3::csp] - energy_cap[region1-2::csp] - energy_cap[region1-1::csp] :   0.0'

        # FIXME: capture Pyomo's print output...
        # assert constraint_string in model._backend_model.reserve_margin_constraint.pprint()

        assert float(model.results.cost.sum()) == approx(282487.35489)



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

energy_balance.py
~~~~~~~~~~~~~~~~~

Energy balance constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import \
    get_param, \
    get_previous_timestep, \
    get_loc_tech_carriers, \
    loc_tech_is_in


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data__['sets']

    if 'loc_carriers_system_balance_constraint' in sets:
        backend_model.system_balance = po.Expression(
            backend_model.loc_carriers_system_balance_constraint,
            backend_model.timesteps,
            initialize=0.0
        )

        backend_model.system_balance_constraint = po.Constraint(
            backend_model.loc_carriers_system_balance_constraint,
            backend_model.timesteps,
            rule=system_balance_constraint_rule
        )

    if 'loc_techs_balance_supply_constraint' in sets:
        backend_model.balance_supply_constraint = po.Constraint(
            backend_model.loc_techs_balance_supply_constraint,
            backend_model.timesteps,
            rule=balance_supply_constraint_rule
        )

    if 'loc_techs_balance_demand_constraint' in sets:
        backend_model.balance_demand_constraint = po.Constraint(
            backend_model.loc_techs_balance_demand_constraint,
            backend_model.timesteps,
            rule=balance_demand_constraint_rule
        )

    if 'loc_techs_balance_transmission_constraint' in sets:
        backend_model.balance_transmission_constraint = po.Constraint(
            backend_model.loc_techs_balance_transmission_constraint,
            backend_model.timesteps,
            rule=balance_transmission_constraint_rule
        )

    if 'loc_techs_resource_availability_supply_plus_constraint' in sets:
        backend_model.balance_supply_plus_constraint = po.Constraint(
            backend_model.loc_techs_resource_availability_supply_plus_constraint,
            backend_model.timesteps,
            rule=balance_supply_plus_constraint_rule
        )

    if 'loc_techs_balance_supply_plus_constraint' in sets:
        backend_model.resource_availability_supply_plus_constraint = po.Constraint(
            backend_model.loc_techs_balance_supply_plus_constraint,
            backend_model.timesteps,
            rule=resource_availability_supply_plus_constraint_rule
        )

    if 'loc_techs_balance_storage_constraint' in sets:
        backend_model.balance_storage_constraint = po.Constraint(
            backend_model.loc_techs_balance_storage_constraint,
            backend_model.timesteps,
            rule=balance_storage_constraint_rule
        )

    if 'loc_techs_balance_storage_inter_cluster_constraint' in sets:
        backend_model.balance_storage_inter_cluster_constraint = po.Constraint(
            backend_model.loc_techs_balance_storage_inter_cluster_constraint,
            backend_model.datesteps,
            rule=balance_storage_inter_cluster_rule
        )

    if 'loc_techs_storage_initial_constraint' in sets:
        backend_model.storage_initial_constraint = po.Constraint(
            backend_model.loc_techs_storage_initial_constraint,
            rule=storage_initial_rule
        )


def system_balance_constraint_rule(backend_model, loc_carrier, timestep):
    """
    System balance ensures that, within each location, the production and
    consumption of each carrier is balanced.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier_{prod} \\in loc::carrier} \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            + \\sum_{loc::tech::carrier_{con} \\in loc::carrier} \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            + \\sum_{loc::tech::carrier_{export} \\in loc::carrier} \\boldsymbol{carrier_{export}}(loc::tech::carrier, timestep)
            \\quad \\forall loc::carrier \\in loc::carriers, \\forall timestep \\in timesteps

    """
    prod, con, export = get_loc_tech_carriers(backend_model, loc_carrier)
    if hasattr(backend_model, 'unmet_demand'):
        unmet_demand = backend_model.unmet_demand[loc_carrier, timestep]
    else:
        unmet_demand = 0

    backend_model.system_balance[loc_carrier, timestep].expr = (
        sum(backend_model.carrier_prod[loc_tech_carrier, timestep] for loc_tech_carrier in prod) +
        sum(backend_model.carrier_con[loc_tech_carrier, timestep] for loc_tech_carrier in con) +
        unmet_demand
    )

    return backend_model.system_balance[loc_carrier, timestep] == 0


def balance_supply_constraint_rule(backend_model, loc_tech, timestep):
    """
    Limit production from supply techs to their available resource

    .. container:: scrolling-wrapper

        .. math::

            min\_use(loc::tech) \\times available\_resource(loc::tech, timestep) \\leq
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep)}
            \\geq available\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{supply}, \\forall timestep \\in timesteps

    If :math:`force\_resource(loc::tech)` is set:

    .. container:: scrolling-wrapper

        .. math::

            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep)}
            = available\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{supply}, \\forall timestep \\in timesteps

    Where:

    .. container:: scrolling-wrapper

        .. math::

            available\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource\_scale(loc::tech)

    if :math:`loc::tech` is in :math:`loc::techs_{area}`:

    .. container:: scrolling-wrapper

        .. math::

            available\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource\_scale(loc::tech) \\times \\boldsymbol{resource_{area}}(loc::tech)

    """
    model_data_dict = backend_model.__calliope_model_data__['data']

    resource = get_param(backend_model, 'resource', (loc_tech, timestep))
    energy_eff = get_param(backend_model, 'energy_eff', (loc_tech, timestep))
    resource_scale = get_param(backend_model, 'resource_scale', loc_tech)
    force_resource = get_param(backend_model, 'force_resource', loc_tech)
    loc_tech_carrier = model_data_dict['lookup_loc_techs'][loc_tech]
    min_use = get_param(backend_model, 'resource_min_use', (loc_tech, timestep))

    if po.value(energy_eff) == 0:
        return backend_model.carrier_prod[loc_tech_carrier, timestep] == 0
    else:
        carrier_prod = backend_model.carrier_prod[loc_tech_carrier, timestep] / energy_eff

    if loc_tech_is_in(backend_model, loc_tech, 'loc_techs_area'):
        available_resource = resource * resource_scale * backend_model.resource_area[loc_tech]
    else:
        available_resource = resource * resource_scale

    if po.value(force_resource):
        return carrier_prod == available_resource
    elif min_use:
        return min_use * available_resource <= carrier_prod <= available_resource
    else:
        return carrier_prod <= available_resource


def balance_demand_constraint_rule(backend_model, loc_tech, timestep):
    """
    Limit consumption from demand techs to their required resource.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep) \\times \\eta_{energy}(loc::tech, timestep) \\geq
            required\_resource(loc::tech, timestep) \\quad \\forall loc::tech \\in loc::techs_{demand},
            \\forall timestep \\in timesteps

    If :math:`force\_resource(loc::tech)` is set:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep) \\times \\eta_{energy}(loc::tech, timestep) =
            required\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{demand}, \\forall timestep \\in timesteps

    Where:

    .. container:: scrolling-wrapper

        .. math::

            required\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource\_scale(loc::tech)

    if :math:`loc::tech` is in :math:`loc::techs_{area}`:

    .. container:: scrolling-wrapper

        .. math::

            required\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource\_scale(loc::tech) \\times \\boldsymbol{resource_{area}}(loc::tech)

    """
    model_data_dict = backend_model.__calliope_model_data__['data']

    resource = get_param(backend_model, 'resource', (loc_tech, timestep))
    energy_eff = get_param(backend_model, 'energy_eff', (loc_tech, timestep))
    resource_scale = get_param(backend_model, 'resource_scale', loc_tech)
    force_resource = get_param(backend_model, 'force_resource', loc_tech)

    loc_tech_carrier = model_data_dict['lookup_loc_techs'][loc_tech]
    carrier_con = backend_model.carrier_con[loc_tech_carrier, timestep] * energy_eff

    if loc_tech_is_in(backend_model, loc_tech, 'loc_techs_area'):
        required_resource = resource * resource_scale * backend_model.resource_area[loc_tech]
    else:
        required_resource = resource * resource_scale

    if po.value(force_resource):
        return carrier_con == required_resource
    else:
        return carrier_con >= required_resource


def resource_availability_supply_plus_constraint_rule(backend_model, loc_tech, timestep):
    """
    Limit production from supply_plus techs to their available resource.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{con}}(loc::tech, timestep)
            \\leq available\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{supply^{+}}, \\forall timestep \\in timesteps

    If :math:`force\_resource(loc::tech)` is set:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{con}}(loc::tech, timestep)
            = available\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{supply^{+}}, \\forall timestep \\in timesteps

    Where:

    .. container:: scrolling-wrapper

        .. math::

            available\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource_{scale}(loc::tech)

    if :math:`loc::tech` is in :math:`loc::techs_{area}`:

    .. container:: scrolling-wrapper

        .. math::

            available\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource_{scale}(loc::tech)
            \\times resource_{area}(loc::tech)

    """
    resource = get_param(backend_model, 'resource', (loc_tech, timestep))
    resource_scale = get_param(backend_model, 'resource_scale', loc_tech)
    force_resource = get_param(backend_model, 'force_resource', loc_tech)

    if loc_tech_is_in(backend_model, loc_tech, 'loc_techs_area'):
        available_resource = resource * resource_scale * backend_model.resource_area[loc_tech]
    else:
        available_resource = resource * resource_scale

    if po.value(force_resource):
        return backend_model.resource_con[loc_tech, timestep] == available_resource
    else:
        return backend_model.resource_con[loc_tech, timestep] <= available_resource


def balance_transmission_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance carrier production and consumption of transmission technologies

    .. container:: scrolling-wrapper

        .. math::

            -1 * \\boldsymbol{carrier_{con}}(loc_{from}::tech:loc_{to}::carrier, timestep)
            \\times \\eta_{energy}(loc::tech, timestep)
            = \\boldsymbol{carrier_{prod}}(loc_{to}::tech:loc_{from}::carrier, timestep)
            \\quad \\forall loc::tech:loc \in locs::techs:locs_{transmission},
            \\forall timestep \in timesteps

    Where a link is the connection between :math:`loc_{from}::tech:loc_{to}`
    and :math:`loc_{to}::tech:loc_{from}` for locations `to` and `from`.

    """
    model_data_dict = backend_model.__calliope_model_data__['data']

    energy_eff = get_param(backend_model, 'energy_eff', (loc_tech, timestep))
    loc_tech_carrier = model_data_dict['lookup_loc_techs'][loc_tech]
    remote_loc_tech = model_data_dict['lookup_remotes'][loc_tech]
    remote_loc_tech_carrier = model_data_dict['lookup_loc_techs'][remote_loc_tech]

    if (loc_tech_is_in(backend_model, remote_loc_tech, 'loc_techs_transmission')
            and get_param(backend_model, 'energy_prod', (loc_tech)) == 1):
        return (
            backend_model.carrier_prod[loc_tech_carrier, timestep] ==
            -1 * backend_model.carrier_con[remote_loc_tech_carrier, timestep] *
            energy_eff
        )
    else:
        return po.Constraint.NoConstraint


def balance_supply_plus_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance carrier production and resource consumption of supply_plus technologies
    alongside any use of resource storage.

    .. _system_balance_constraint:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep) =
            \\boldsymbol{storage}(loc::tech, timestep_{previous})
            \\times (1 - storage\_loss(loc::tech, timestep))^{timestep\_resolution(timestep)} +
            \\boldsymbol{resource_{con}}(loc::tech, timestep) \\times \\eta_{resource}(loc::tech, timestep) -
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep) \\times \\eta_{parasitic}(loc::tech, timestep)}
            \\quad \\forall loc::tech \\in loc::techs_{supply^{+}}, \\forall timestep \\in timesteps

    If *no* storage is defined for the technology, this reduces to:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{con}}(loc::tech, timestep) \\times \\eta_{resource}(loc::tech, timestep) =
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep) \\times \\eta_{parasitic}(loc::tech, timestep)}
            \\quad \\forall loc::tech \\in loc::techs_{supply^{+}}, \\forall timestep \\in timesteps

    """

    model_data_dict = backend_model.__calliope_model_data__['data']
    model_attrs = backend_model.__calliope_model_data__['attrs']

    resource_eff = get_param(backend_model, 'resource_eff', (loc_tech, timestep))
    energy_eff = get_param(backend_model, 'energy_eff', (loc_tech, timestep))
    parasitic_eff = get_param(backend_model, 'parasitic_eff', (loc_tech, timestep))
    total_eff = energy_eff * parasitic_eff

    if po.value(total_eff) == 0:
        carrier_prod = 0
    else:
        loc_tech_carrier = model_data_dict['lookup_loc_techs'][loc_tech]
        carrier_prod = backend_model.carrier_prod[loc_tech_carrier, timestep] / total_eff

    # A) Case where no storage allowed
    if not loc_tech_is_in(backend_model, loc_tech, 'loc_techs_store'):
        return backend_model.resource_con[loc_tech, timestep] * resource_eff == carrier_prod

    # B) Case where storage is allowed
    else:
        resource = backend_model.resource_con[loc_tech, timestep] * resource_eff
        current_timestep = backend_model.timesteps.order_dict[timestep]
        if current_timestep == 0 and not model_attrs['run.cyclic_storage']:
            storage_previous_step = get_param(backend_model, 'storage_initial', loc_tech)
        elif (hasattr(backend_model, 'storage_inter_cluster') and
                model_data_dict['lookup_cluster_first_timestep'][timestep]):
            storage_previous_step = 0
        else:
            if (hasattr(backend_model, 'clusters') and
                    model_data_dict['lookup_cluster_first_timestep'][timestep]):
                previous_step = model_data_dict['lookup_cluster_last_timestep'][timestep]
            elif current_timestep == 0 and model_attrs['run.cyclic_storage']:
                previous_step = backend_model.timesteps[-1]
            else:
                previous_step = get_previous_timestep(backend_model.timesteps, timestep)
            storage_loss = get_param(backend_model, 'storage_loss', loc_tech)
            time_resolution = backend_model.timestep_resolution[previous_step]
            storage_previous_step = (
                ((1 - storage_loss) ** time_resolution) *
                backend_model.storage[loc_tech, previous_step]
            )

        return (
            backend_model.storage[loc_tech, timestep] ==
            storage_previous_step + resource - carrier_prod
        )


def balance_storage_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance carrier production and consumption of storage technologies,
    alongside any use of the stored volume.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep) =
            \\boldsymbol{storage}(loc::tech, timestep_{previous})
            \\times (1 - storage\_loss(loc::tech, timestep))^{resolution(timestep)}
            - \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            \\times \\eta_{energy}(loc::tech, timestep)
            - \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep)}
            \\quad \\forall loc::tech \\in loc::techs_{storage}, \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data__['data']
    model_attrs = backend_model.__calliope_model_data__['attrs']

    energy_eff = get_param(backend_model, 'energy_eff', (loc_tech, timestep))

    if po.value(energy_eff) == 0:
        carrier_prod = 0
    else:
        loc_tech_carrier = model_data_dict['lookup_loc_techs'][loc_tech]
        carrier_prod = backend_model.carrier_prod[loc_tech_carrier, timestep] / energy_eff

    carrier_con = backend_model.carrier_con[loc_tech_carrier, timestep] * energy_eff

    current_timestep = backend_model.timesteps.order_dict[timestep]
    if current_timestep == 0 and not model_attrs['run.cyclic_storage']:
        storage_previous_step = get_param(backend_model, 'storage_initial', loc_tech)
    elif (hasattr(backend_model, 'storage_inter_cluster') and
            model_data_dict['lookup_cluster_first_timestep'][timestep]):
        storage_previous_step = 0
    else:
        if (hasattr(backend_model, 'clusters') and
                model_data_dict['lookup_cluster_first_timestep'][timestep]):
            previous_step = model_data_dict['lookup_cluster_last_timestep'][timestep]
        elif current_timestep == 0 and model_attrs['run.cyclic_storage']:
            previous_step = backend_model.timesteps[-1]
        else:
            previous_step = get_previous_timestep(backend_model.timesteps, timestep)
        storage_loss = get_param(backend_model, 'storage_loss', loc_tech)
        time_resolution = backend_model.timestep_resolution[previous_step]
        storage_previous_step = (
            ((1 - storage_loss) ** time_resolution) *
            backend_model.storage[loc_tech, previous_step]
        )

    return (
        backend_model.storage[loc_tech, timestep] ==
        storage_previous_step - carrier_prod - carrier_con
    )


def balance_storage_inter_cluster_rule(backend_model, loc_tech, datestep):
    """
    When clustering days, to reduce the timeseries length, balance the daily stored
    energy across all days of the original timeseries.

    `Ref: DOI 10.1016/j.apenergy.2018.01.023 <https://doi.org/10.1016/j.apenergy.2018.01.023>`_

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{inter\_cluster}}(loc::tech, datestep) =
            \\boldsymbol{storage_{inter\_cluster}}(loc::tech, datestep_{previous})
            \\times (1 - storage\_loss(loc::tech, timestep))^{24}
            + \\boldsymbol{storage}(loc::tech, timestep_{final, cluster(datestep))})
            \\quad \\forall loc::tech \\in loc::techs_{store}, \\forall datestep \\in datesteps

    Where :math:`timestep_{final, cluster(datestep_{previous}))}` is the final timestep of the
    cluster in the clustered timeseries corresponding to the previous day
    """
    model_attrs = backend_model.__calliope_model_data__['attrs']
    current_datestep = backend_model.datesteps.order_dict[datestep]

    if current_datestep == 0 and not model_attrs['run.cyclic_storage']:
        storage_previous_step = get_param(backend_model, 'storage_initial', loc_tech)
        storage_intra = 0
    else:
        if current_datestep == 0 and model_attrs['run.cyclic_storage']:
            previous_step = backend_model.datesteps[-1]
        else:
            previous_step = get_previous_timestep(backend_model.datesteps, datestep)
        storage_loss = get_param(backend_model, 'storage_loss', loc_tech)
        storage_previous_step = (
            ((1 - storage_loss) ** 24) *
            backend_model.storage_inter_cluster[loc_tech, previous_step]
        )
        final_timestep = (
            backend_model.__calliope_model_data__
            ['data']['lookup_datestep_last_cluster_timestep'][previous_step]
        )
        storage_intra = backend_model.storage[loc_tech, final_timestep]
    return (
        backend_model.storage_inter_cluster[loc_tech, datestep] ==
        storage_previous_step + storage_intra
    )


def storage_initial_rule(backend_model, loc_tech):
    """
    If storage is cyclic, allow an initial storage to still be set. This is
    applied to the storage of the final timestep/datestep of the series as that,
    in cyclic storage, is the 'storage_previous_step' for the first
    timestep/datestep.

    If clustering and ``storage_inter_cluster`` exists:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{inter\_cluster}}(loc::tech, datestep_{final})
            \\times ((1 - storage_loss) ** 24) = storage_{initial}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{store}, \\forall datestep \\in datesteps

    Where :math:`datestep_{final}` is the last datestep of the timeseries

    Else:
    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep_{final})
            \\times ((1 - storage_loss) ** 24) = storage_{initial}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{store}, \\forall timestep \\in timesteps

    Where :math:`timestep_{final}` is the last timestep of the timeseries
    """

    storage_initial = get_param(backend_model, 'storage_initial', loc_tech)
    storage_loss = get_param(backend_model, 'storage_loss', loc_tech)
    if hasattr(backend_model, 'storage_inter_cluster'):
        storage = backend_model.storage_inter_cluster
        final_step = backend_model.datesteps[-1]
        time_resolution = 24
    else:
        storage = backend_model.storage
        final_step = backend_model.timesteps[-1]
        time_resolution = backend_model.timestep_resolution[final_step]

    return (
        storage[loc_tech, final_step] * ((1 - storage_loss) ** time_resolution)
        == storage_initial
    )



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

run_checks.py
~~~~~~~~~~~~~

Checks for model consistency and possible errors when preparing run in the backend.

"""
import ruamel.yaml

import numpy as np
import xarray as xr
from calliope.core.attrdict import AttrDict


def check_operate_params(model_data):
    """
    if model mode = `operate`, check for clashes in capacity constraints.
    In this mode, all capacity constraints are set to parameters in the backend,
    so can easily lead to model infeasibility if not checked.

    Returns
    -------
    comments : AttrDict
        debug output
    warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    defaults = ruamel.yaml.load(model_data.attrs['defaults'], Loader=ruamel.yaml.Loader)
    warnings, errors = [], []
    comments = AttrDict()

    def _get_param(loc_tech, var):
        if _is_in(loc_tech, var) and not np.isnan(model_data[var].loc[loc_tech].item()):
            param = model_data[var].loc[loc_tech].item()
        else:
            param = defaults[var]
        return param

    def _is_in(loc_tech, set_or_var):
        if set_or_var in model_data:
            try:
                model_data[set_or_var].loc[loc_tech]
                return True
            except KeyError:
                return False
        else:
            return False

    for loc_tech in model_data.loc_techs.values:
        energy_cap = model_data.energy_cap.loc[loc_tech].item()
        # Must have energy_cap defined for all relevant techs in the model
        if (np.isnan(energy_cap) or np.isinf(energy_cap)) and not _is_in(loc_tech, 'force_resource'):
            errors.append(
                'Operate mode: User must define a finite energy_cap (via '
                'energy_cap_equals or energy_cap_max) for {}'.format(loc_tech)
            )

        elif _is_in(loc_tech, 'loc_techs_finite_resource'):
            # force resource overrides capacity constraints, so set capacity constraints to infinity
            if _is_in(loc_tech, 'force_resource'):
                energy_cap = model_data.energy_cap.loc[loc_tech] = np.inf
                warnings.append(
                    'Energy capacity constraint removed from {} as force_resource '
                    'is applied'.format(loc_tech)
                )
                if _is_in(loc_tech, 'resource_cap'):
                    resource_cap = model_data.resource_cap.loc[loc_tech] = np.inf
                    warnings.append(
                        'Resource capacity constraint removed from {} as force_resource '
                        'is applied'.format(loc_tech)
                    )
            # Cannot have infinite resource area (physically impossible)
            if _is_in(loc_tech, 'loc_techs_area'):
                area = model_data.resource_area.loc[loc_tech].item()
                if np.isnan(area) or np.isinf(area):
                    errors.append(
                        'Operate mode: User must define a finite resource_area '
                        '(via resource_area_equals or resource_area_max) for {}, '
                        'as a finite available resource is considered'.format(loc_tech)
                    )
            # Cannot have consumed resource being higher than energy_cap, as
            # constraints will clash. Doesn't affect supply_plus techs with a
            # storage buffer prior to carrier production.
            elif not _is_in(loc_tech, 'loc_techs_store'):
                resource_scale = _get_param(loc_tech, 'resource_scale')
                energy_cap_scale = _get_param(loc_tech, 'energy_cap_scale')
                resource_eff = _get_param(loc_tech, 'resource_eff')
                energy_eff = _get_param(loc_tech, 'energy_eff')
                energy_cap_scale = _get_param(loc_tech, 'energy_cap_scale')
                resource = model_data.resource.loc[loc_tech].values
                if (energy_cap is not None and
                    any(resource * resource_scale * resource_eff >
                        energy_cap * energy_cap_scale * energy_eff)):
                    errors.append(
                        'Operate mode: resource is forced to be higher than '
                        'fixed energy cap for `{}`'.format(loc_tech)
                    )
        # Must define a resource capacity to ensure the Pyomo param is created
        # for it. But we just create an array of infs, so the capacity has no effect
        if _is_in(loc_tech, 'loc_techs_supply_plus'):
            if 'resource_cap' not in model_data.data_vars.keys():
                model_data['resource_cap'] = xr.DataArray(
                    [np.inf for i in model_data.loc_techs_supply_plus.values],
                    dims='loc_techs_supply_plus')
                model_data['resource_cap'].attrs['is_result'] = 1
                model_data['resource_cap'].attrs['operate_param'] = 1
                warnings.append(
                    'Resource capacity constraint defined and set to infinity '
                    'for all supply_plus techs'
                )

        if _is_in(loc_tech, 'loc_techs_store'):
            if _is_in(loc_tech, 'charge_rate'):
                storage_cap = model_data.storage_cap.loc[loc_tech].item()
                if storage_cap and energy_cap:
                    charge_rate = model_data['charge_rate'].loc[loc_tech]
                    if storage_cap * charge_rate < energy_cap:
                        errors.append(
                            'fixed storage capacity * charge rate is not larger '
                            'than fixed energy capacity for loc::tech {}'.format(loc_tech)
                        )

    window = model_data.attrs.get('run.operation.window', None)
    horizon = model_data.attrs.get('run.operation.horizon', None)
    if not window or not horizon:
        errors.append(
            'Operational mode requires a timestep window and horizon to be '
            'defined under run.operation'
        )
    elif horizon < window:
        errors.append(
            'Iteration horizon must be larger than iteration window, '
            'for operational mode'
        )

    # Cyclic storage isn't really valid in operate mode, so we ignore it, using
    # initial_storage instead (allowing us to pass storage between operation windows)
    # TODO: update default to True for 0.6.3
    if model_data.attrs.get('run.cyclic_storage', False):
        warnings.append(
            'Storage cannot be cyclic in operate run mode, setting '
            '`run.cyclic_storage` to False for this run'
        )
        model_data.attrs['run.cyclic_storage'] = False

    return comments, warnings, errors



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

cli.py
~~~~~~

Command-line interface.

"""

import contextlib
import datetime
import logging
import os
import pstats
import shutil
import sys
import traceback
import click

from calliope import Model, read_netcdf, examples
from calliope.core.util.convert import convert_model
from calliope.core.util.generate_runs import generate
from calliope.core.util.logging import logger, set_log_level
from calliope._version import __version__


_time_format = '%Y-%m-%d %H:%M:%S'


_debug = click.option(
    '--debug', is_flag=True, default=False,
    help='Print debug information when encountering errors.'
)

_quiet = click.option(
    '--quiet', is_flag=True, default=False,
    help='Be less verbose about what is happening, including hiding '
         'solver output.'
)

_pdb = click.option(
    '--pdb', is_flag=True, default=False,
    help='If used together with --debug, drop into interactive '
         'debugger on encountering errors.'
)

_profile = click.option(
    '--profile', is_flag=True, default=False,
    help='Run through cProfile.'
)

_profile_filename = click.option(
    '--profile_filename', type=str,
    help='Filename to save profile to if enabled --profile.'
)

if logger.hasHandlers():
    for handler in logger.handlers:
        logger.removeHandler(handler)

formatter = logging.Formatter(
    '[%(asctime)s] %(levelname)-8s %(message)s', datefmt=_time_format
)
console = logging.StreamHandler(stream=sys.stderr)
console.setFormatter(formatter)
logger.addHandler(console)

@contextlib.contextmanager
def format_exceptions(debug=False, pdb=False, profile=False, profile_filename=None, start_time=None):
    if debug:
        set_log_level('DEBUG')

    try:
        if profile:
            import cProfile
            profile = cProfile.Profile()
            profile.enable()
        yield
        if profile:
            profile.disable()
            if profile_filename:
                dump_path = os.path.expanduser(profile_filename)
                print('\nSaving cProfile output to: {}'.format(dump_path))
                profile.dump_stats(dump_path)
            else:
                print('\n\n----PROFILE OUTPUT----\n\n')
                stats = pstats.Stats(profile).sort_stats('cumulative')
                stats.print_stats(20)  # Print first 20 lines

    except Exception as e:
        if debug:
            traceback.print_exc()
            if pdb:
                import ipdb
                ipdb.post_mortem(e.__traceback__)
        else:
            stack = traceback.extract_tb(e.__traceback__)
            # Get last stack trace entry still in Calliope
            last = [i for i in stack if 'calliope' in i[0]][-1]
            if debug:
                err_string = '\nError in {}, {}:{}'.format(last[2], last[0], last[1])
            else:
                err_string = '\nError in {}:'.format(last[2])
            click.secho(err_string, fg='red')
            click.secho(str(e), fg='red')
            if start_time:
                print_end_time(start_time, msg='aborted due to an error')
        sys.exit(1)


def set_quietness_level(quiet):
    if quiet:
        set_log_level('WARNING')
    else:
        set_log_level('SOLVER')


def print_end_time(start_time, msg='complete'):
    end_time = datetime.datetime.now()
    secs = round((end_time - start_time).total_seconds(), 1)
    tend = end_time.strftime(_time_format)
    print('\nCalliope run {}. '
          'Elapsed: {} seconds (time at exit: {})'.format(msg, secs, tend))


def _get_version():
    return 'Version {}'.format(__version__)


@click.group(invoke_without_command=True)
@click.pass_context
@click.option('--version', is_flag=True, default=False,
              help='Display version.')
def cli(ctx, version):
    """Calliope: a multi-scale energy systems modelling framework"""
    if ctx.invoked_subcommand is None and not version:
        print(ctx.get_help())
    if version:
        print(_get_version())


@cli.command(short_help='Create a new model based on a built-in example.')
@click.argument('path')
@click.option('--template', type=str, default=None)
@_debug
def new(path, template, debug):
    """
    Create new model at the given ``path``, based on one of the built-in
    example models. The target path must not yet exist. Intermediate
    directories will be created automatically.
    """
    if debug:
        print(_get_version())
    with format_exceptions(debug):
        if template is None:
            template = 'national_scale'
        source_path = examples._PATHS[template]
        click.echo('Copying {} template to target directory: {}'.format(template, path))
        shutil.copytree(source_path, path)


@cli.command(short_help='Run a model.')
@click.argument('model_file')
@click.option('--override_file')
@click.option('--save_netcdf')
@click.option('--save_csv')
@click.option('--save_plots')
@click.option('--save_logs')
@click.option('--model_format')
@_debug
@_quiet
@_pdb
@_profile
@_profile_filename
def run(model_file, override_file, save_netcdf, save_csv, save_plots,
        save_logs, model_format,
        debug, quiet, pdb, profile, profile_filename):
    """
    Execute the given model. Tries to guess from the file extension whether
    ``model_file`` is a YAML file or a pre-built model saved to NetCDF.
    This can also explicitly be set with the --model_format=yaml or
    --model_format=netcdf option.

    """
    if debug:
        print(_get_version())

    set_quietness_level(quiet)

    logging.captureWarnings(True)
    pywarning_logger = logging.getLogger('py.warnings')
    pywarning_logger.addHandler(console)

    start_time = datetime.datetime.now()
    with format_exceptions(debug, pdb, profile, profile_filename, start_time):
        if save_csv is None and save_netcdf is None:
            click.secho(
                '\n!!!\nWARNING: No options to save results have been '
                'specified.\nModel will run without saving results!\n!!!\n',
                fg='red', bold=True
            )
        tstart = start_time.strftime(_time_format)
        print('Calliope {} starting at {}\n'.format(__version__, tstart))

        # Try to determine model file type if not given explicitly
        if model_format is None:
            if model_file.split('.')[-1] in ['yaml', 'yml']:
                model_format = 'yaml'
            elif model_file.split('.')[-1] in ['nc', 'nc4', 'netcdf']:
                model_format = 'netcdf'
            else:
                raise ValueError(
                    'Cannot determine model file format based on file '
                    'extension for "{}". Set format explicitly with '
                    '--model_format.'.format(model_file)
                )

        if model_format == 'yaml':
            override_dict = {
                'run.save_logs': save_logs
            }
            model = Model(
                model_file, override_file=override_file, override_dict=override_dict
            )
        elif model_format == 'netcdf':
            if override_file is not None:
                raise ValueError(
                    'Overrides cannot be applied when loading a pre-built '
                    'model from NetCDF. Please run without --override options.'
                )
            model = read_netcdf(model_file)
            if save_logs is not None:
                model._model_data.attrs['run.save_logs'] = save_logs
        else:
            raise ValueError('Invalid model format: {}'.format(model_format))

        print(model.info() + '\n')
        print('Starting model run...')
        model.run()

        termination = model._model_data.attrs.get(
            'termination_condition', 'unknown')
        if save_csv:
            print('Saving CSV results to directory: {}'.format(save_csv))
            model.to_csv(save_csv)
        if save_netcdf:
            print('Saving NetCDF results to file: {}'.format(save_netcdf))
            model.to_netcdf(save_netcdf)
        if save_plots:
            if termination == 'optimal':
                print('Saving HTML file with plots to: {}'.format(save_plots))
                model.plot.summary(to_file=save_plots)
            else:
                click.secho(
                    'Model termination condition non-optimal. Not saving plots',
                    fg='red', bold=True
                )
        print_end_time(start_time)


@cli.command(short_help='Generate a script to run multiple models.')
@click.argument('model_file')
@click.argument('out_file')
@click.option('--kind', help='One of: "bash", "bsub".')
@click.option('--override_file')
@click.option('--groups')
@click.option('--cluster_threads', default=1)
@click.option('--cluster_mem')
@click.option('--cluster_time')
@click.option(
    '--additional_args', default='',
    help='Any additional arguments to pass directly on to `calliope run`.')
@_debug
@_quiet
@_pdb
def generate_runs(
        model_file, out_file, kind, override_file, groups, additional_args,
        cluster_threads, cluster_mem, cluster_time,
        debug, quiet, pdb):

    set_quietness_level(quiet)

    kwargs = dict(
        model_file=model_file,
        out_file=out_file,
        override_file=override_file,
        groups=groups,
        additional_args=additional_args,
        cluster_mem=cluster_mem,
        cluster_time=cluster_time,
        cluster_threads=cluster_threads,
    )

    with format_exceptions(debug, pdb):
        generate(kind, **kwargs)


@cli.command(short_help='Convert a 0.5.x model to 0.6.0.')
@click.argument('run_config_path')
@click.argument('model_config_path')
@click.argument('out_path')
@_debug
@_quiet
@_pdb
def convert(
        run_config_path, model_config_path, out_path,
        debug, quiet, pdb):
    """
    Convert a Calliope 0.5.x model specified by its run and model
    configurations to a Calliope 0.6.0 model.

    The input model is not modified; the converted version of the model
    is saved to ``out_path``.

    """
    set_quietness_level(quiet)
    with format_exceptions(debug, pdb):
        print('Converting model...')
        convert_model(run_config_path, model_config_path, out_path)
        print('Done. Results saved to: {}'.format(out_path))



from pytest import approx

import calliope

_OVERRIDE_FILE = calliope.examples._PATHS['national_scale'] + '/overrides.yaml'


class TestNationalScaleObjectives:
    def test_nationalscale_minimize_emissions(self):
        model = calliope.examples.national_scale(
            override_file=_OVERRIDE_FILE + ':minimize_emissions_costs',
            override_dict={
                'model.subset_time': '2005-01-01'
            }
        )
        model.run()

        assert model.results.energy_cap.to_pandas()['region1-2::csp'] == approx(10000.0)
        assert model.results.energy_cap.to_pandas()['region1::ac_transmission:region2'] == approx(3423.336471)

        assert model.results.carrier_prod.sum('timesteps').to_pandas()['region1::ccgt::power'] == approx(272792.633882353)

        assert float(model.results.cost.sum()) == approx(32039968.579169072)

    def test_nationalscale_maximize_utility(self):
        model = calliope.examples.national_scale(
            override_file=_OVERRIDE_FILE + ':maximize_utility_costs',
            override_dict={
                'model.subset_time': '2005-01-01'
            }
        )
        model.run()

        assert model.results.energy_cap.to_pandas()['region1-2::csp'] == approx(10000.0)
        assert model.results.energy_cap.to_pandas()['region1::ac_transmission:region2'] == approx(10000.0)

        assert model.results.carrier_prod.sum('timesteps').to_pandas()['region1::ccgt::power'] == approx(340154.9722)

        assert float(model.results.cost.sum()) == approx(47857169.51379299)



import pytest  # pylint: disable=unused-import
import numpy as np
import xarray as xr
import pyomo.core as po
from pyomo.core.base.expr import identify_variables

from calliope.backend.pyomo.util import get_param
import calliope.exceptions as exceptions

from calliope.test.common.util import build_test_model as build_model
from calliope.test.common.util import check_error_or_warning


def check_variable_exists(backend_model, constraint, variable):
    """
    Search for existence of a decision variable in a Pyomo constraint.

    Parameters
    ----------
    backend_model : Pyomo ConcreteModel
    constraint : str, name of constraint which could exist in the backend
    variable : str, string to search in the list of variables to check if existing
    """
    exists = []
    for v in getattr(backend_model, constraint).values():
        variables = identify_variables(v.body)
        exists.append(any(variable in j.getname() for j in list(variables)))
    return any(exists)


def check_standard_warning(info, warning):

    if warning == 'transmission':
        return check_error_or_warning(
            info,
            'dimension loc_techs_transmission and associated variables distance, '
            'lookup_remotes were empty, so have been deleted'
        )


class TestUtil:
    def test_get_param_with_timestep_existing(self):
        """
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run()
        param = get_param(
            m._backend_model,
            'resource',
            ('1::test_demand_elec', m._backend_model.timesteps[1])
        )
        assert po.value(param) == -5  # see demand_elec.csv

    def test_get_param_no_timestep_existing(self):
        """
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run()
        param = get_param(
            m._backend_model,
            'energy_eff',
            ('1::test_supply_elec', m._backend_model.timesteps[1])
        )
        assert po.value(param) == 0.9  # see test model.yaml

    def test_get_param_no_timestep_possible(self):
        """
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run()
        param = get_param(
            m._backend_model,
            'energy_cap_max',
            ('1::test_supply_elec')
        )
        assert po.value(param) == 10  # see test model.yaml

        param = get_param(
            m._backend_model,
            'cost_energy_cap',
            ('monetary', '0::test_supply_elec')
        )
        assert po.value(param) == 10

    def test_get_param_from_default(self):
        """
        """
        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run()

        param = get_param(
            m._backend_model,
            'parasitic_eff',
            ('1::test_supply_plus', m._backend_model.timesteps[1])
        )
        assert po.value(param) == 1  # see defaults.yaml

        param = get_param(
            m._backend_model,
            'resource_cap_min',
            ('0::test_supply_plus')
        )
        assert po.value(param) == 0  # see defaults.

        param = get_param(
            m._backend_model,
            'cost_resource_cap',
            ('monetary', '1::test_supply_plus')
        )
        assert po.value(param) == 0  # see defaults.yaml

    def test_get_param_no_default_defined(self):
        """
        If a default is not defined, raise KeyError
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run()
        with pytest.raises(KeyError):
            get_param(
                m._backend_model,
                'random_param',
                ('1::test_demand_elec', m._backend_model.timesteps[1])
            )
            get_param(
                m._backend_model,
                'random_param',
                ('1::test_supply_elec')
            )


class TestInterface:
    def test_get_input_params(self):
        """
        Test that the function access_model_inputs works
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run()
        m.backend.access_model_inputs()

    def test_update_param(self):
        """
        test that the function update_param works
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run()
        m.backend.update_param('energy_cap_max', '1::test_supply_elec', 20)
        assert (
            m._backend_model.energy_cap_max.extract_values()['1::test_supply_elec'] == 20
        )

    def test_activate_constraint(self):
        """
        test that the function activate_constraint works
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run()
        m.backend.activate_constraint('system_balance_constraint', active=False)
        assert not m._backend_model.system_balance_constraint.active

    def test_rerun(self):
        """
        test that the function rerun works
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run()
        returned_dataset = m.backend.rerun()
        assert isinstance(returned_dataset, xr.Dataset)

        # should fail if the run mode is not 'plan'
        with pytest.raises(exceptions.ModelError) as error:
            m._model_data.attrs['run.mode'] = 'operate'
            m.backend.rerun()
        assert check_error_or_warning(error, 'Cannot rerun the backend in operate run mode')


class TestChecks:
    def test_operate_cyclic_storage(self):
        """Cannot have cyclic storage in operate mode"""
        override = {'run.cyclic_storage': True}
        m = build_model(override, 'simple_supply,operate,investment_costs')
        assert m._model_data.attrs['run.cyclic_storage'] is True
        with pytest.warns(exceptions.ModelWarning) as warning:
            m.run(build_only=True)
        assert check_error_or_warning(warning, 'Storage cannot be cyclic in operate run mode')
        assert m._model_data.attrs['run.cyclic_storage'] is False


class TestBalanceConstraints:

    def test_loc_carriers_system_balance_constraint(self):
        """
        sets.loc_carriers
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'system_balance_constraint')

    def test_loc_techs_balance_supply_constraint(self):
        """
        sets.loc_techs_finite_resource_supply,
        """
        m = build_model({'techs.test_supply_elec.constraints.resource': 20},
                        'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_supply_constraint')

    def test_loc_techs_balance_demand_constraint(self):
        """
        sets.loc_techs_finite_resource_demand,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_demand_constraint')

    def test_loc_techs_resource_availability_supply_plus_constraint(self):
        """
        sets.loc_techs_finite_resource_supply_plus,
        """
        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, 'resource_availability_supply_plus_constraint'
        )

    def test_loc_techs_balance_transmission_constraint(self):
        """
        sets.loc_techs_transmission,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_transmission_constraint')

    def test_loc_techs_balance_supply_plus_constraint(self):
        """
        sets.loc_techs_supply_plus,
        """
        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_supply_plus_constraint')

    def test_loc_techs_balance_storage_constraint(self):
        """
        sets.loc_techs_storage,
        """
        m = build_model({}, 'simple_storage,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_storage_constraint')
        assert not hasattr(m._backend_model, 'storage_initial_constraint')

    def test_storage_initial_constraint(self):
        """
        sets.loc_techs_store,
        """
        m = build_model(
            {'run.cyclic_storage': True},
            'simple_storage,one_day,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_storage_constraint')
        assert not hasattr(m._backend_model, 'storage_initial_constraint')

        m2 = build_model(
            {'run.cyclic_storage': True,
             'techs.test_storage.constraints.storage_initial': 0},
            'simple_storage,one_day,investment_costs'
        )
        m2.run(build_only=True)
        assert hasattr(m2._backend_model, 'balance_storage_constraint')
        assert hasattr(m2._backend_model, 'storage_initial_constraint')

    def test_carriers_reserve_margin_constraint(self):
        """
        i for i in sets.carriers if i in model_run.model.get_key('reserve_margin', {}).keys()
        """
        m = build_model({'model.reserve_margin.electricity': 0.01}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'reserve_margin_constraint')


class TestCostConstraints:
    # costs.py
    def test_loc_techs_cost_constraint(self):
        """
        sets.loc_techs_cost,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_constraint')

    def test_loc_techs_cost_investment_constraint(self):
        """
        sets.loc_techs_investment_cost,
        """
        m = build_model({}, 'simple_conversion,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_investment_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.lifetime': 10,
                'techs.test_supply_elec.costs.monetary.interest_rate': 0.1},
            'supply_purchase,two_hours'
        )
        m.run(build_only=True)

        assert hasattr(m._backend_model, 'cost_investment_constraint')

    def test_loc_techs_cost_var_constraint(self):
        """
        i for i in sets.loc_techs_om_cost if i not in sets.loc_techs_conversion_plus + sets.loc_techs_conversion

        """
        m = build_model({}, 'simple_conversion,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'cost_var_constraint')

        m = build_model(
            {'techs.test_conversion.costs.monetary.om_con': 1},
            'simple_conversion,two_hours'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_rhs')
        assert not hasattr(m._backend_model, 'cost_var_constraint')

        m = build_model(
            {'techs.test_conversion_plus.costs.monetary.om_prod': 1},
            'simple_conversion_plus,two_hours'
        )

        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_rhs')
        assert not hasattr(m._backend_model, 'cost_var_constraint')

        m = build_model(
            {'techs.test_supply_elec.costs.monetary.om_prod': 1},
            'simple_supply,two_hours'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_constraint')

        m = build_model(
            {'techs.test_supply_elec.costs.monetary.om_con': 1},
            'simple_supply,two_hours'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_constraint')

        m = build_model(
            {'techs.test_supply_plus.costs.monetary.om_con': 1},
            'simple_supply_and_supply_plus,two_hours'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_constraint')


class TestExportConstraints:
    # export.py
    def test_loc_carriers_update_system_balance_constraint(self):
        """
        i for i in sets.loc_carriers if sets.loc_techs_export
        and any(['{0}::{2}'.format(*j.split('::')) == i
        for j in sets.loc_tech_carriers_export])
        """

        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        export_exists = check_variable_exists(
            m._backend_model, 'system_balance_constraint', 'carrier_export'
        )
        assert not export_exists

        m = build_model({}, 'supply_export,two_hours,investment_costs')
        m.run(build_only=True)

        export_exists = check_variable_exists(
            m._backend_model, 'system_balance_constraint', 'carrier_export'
        )
        assert export_exists

    def test_loc_tech_carriers_export_balance_constraint(self):
        """
        sets.loc_tech_carriers_export,
        """

        m = build_model({}, 'supply_export,two_hours,investment_costs')
        m.run(build_only=True)

        assert hasattr(m._backend_model, 'export_balance_constraint')

    def test_loc_techs_update_costs_var_constraint(self):
        """
        i for i in sets.loc_techs_om_cost if i in sets.loc_techs_export
        """

        m = build_model({}, 'supply_export,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_rhs')
        assert hasattr(m._backend_model, 'cost_var_constraint')

        m = build_model(
            {'techs.test_supply_elec.costs.monetary.om_prod': 0.1},
            'supply_export,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_rhs')
        assert hasattr(m._backend_model, 'cost_var_constraint')

        export_exists = check_variable_exists(
            m._backend_model, 'cost_var_constraint', 'carrier_export'
        )
        assert export_exists

    def test_loc_tech_carriers_export_max_constraint(self):
        """
        i for i in sets.loc_tech_carriers_export
        if constraint_exists(model_run, i.rsplit('::', 1)[0], 'constraints.export_cap')
        """

        m = build_model(
            {'techs.test_supply_elec.constraints.export_cap': 5},
            'supply_export,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'export_max_constraint')


class TestCapacityConstraints:
    # capacity.py
    def test_loc_techs_storage_capacity_constraint(self):
        """
        i for i in sets.loc_techs_store if i not in sets.loc_techs_milp
        """
        m = build_model({}, 'simple_storage,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'storage_capacity_constraint')

        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'storage_capacity_constraint')

        m = build_model(
            {'techs.test_storage.constraints':
                {'units_max': 1, 'energy_cap_per_unit': 20,
                 'storage_cap_per_unit': 20}},
            'simple_storage,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_constraint')

        m = build_model(
            {'techs.test_storage.constraints.storage_cap_equals': 20},
            'simple_storage,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert m._backend_model.storage_capacity_constraint['0::test_storage'].upper() == 20
        assert m._backend_model.storage_capacity_constraint['0::test_storage'].lower() == 20

    def test_loc_techs_energy_capacity_storage_constraint(self):
        """
        i for i in sets.loc_techs_store if constraint_exists(model_run, i, 'constraints.charge_rate')
        """
        m = build_model({}, 'simple_storage,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_storage_constraint')

        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_storage_constraint')

        # constraint should exist in the MILP case too
        m = build_model({}, 'supply_and_supply_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_storage_constraint')

    def test_loc_techs_resource_capacity_constraint(self):
        """
        i for i in sets.loc_techs_finite_resource_supply_plus
        if any([constraint_exists(model_run, i, 'constraints.resource_cap_equals'),
                constraint_exists(model_run, i, 'constraints.resource_cap_max'),
                constraint_exists(model_run, i, 'constraints.resource_cap_min')])
        """

        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'resource_capacity_constraint')

        m = build_model(
            {'techs.test_supply_plus.constraints.resource_cap_max': 10},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_storage_constraint')

        m = build_model(
            {'techs.test_supply_plus.constraints.resource_cap_min': 10},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_storage_constraint')

        m = build_model(
            {'techs.test_supply_plus.constraints.resource_cap_equals': 10},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_storage_constraint')

    def test_loc_techs_resource_capacity_equals_energy_capacity_constraint(self):
        """
        i for i in sets.loc_techs_finite_resource_supply_plus
        if constraint_exists(model_run, i, 'constraints.resource_cap_equals_energy_cap')
        """

        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, 'resource_capacity_equals_energy_capacity_constraint'
        )

        m = build_model(
            {'techs.test_supply_plus.constraints.resource_cap_equals_energy_cap': True},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, 'resource_capacity_equals_energy_capacity_constraint'
        )

    def test_loc_techs_resource_area_constraint(self):
        """
        i for i in sets.loc_techs_area if i in sets.loc_techs_supply_plus
        """

        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'resource_area_constraint')

        m = build_model(
            {'techs.test_supply_plus.constraints.resource_area_max': 10},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'resource_area_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.resource_area_max': 10},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'resource_area_constraint')

        # Check that setting energy_cap_max to 0 also forces this constraint to 0
        m = build_model(
            {'techs.test_supply_plus.constraints': {'resource_area_max': 10, 'energy_cap_max': 0}},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert m._backend_model.resource_area_constraint['0::test_supply_plus'].upper() == 0

    def test_loc_techs_resource_area_per_energy_capacity_constraint(self):
        """
        i for i in sets.loc_techs_area if i in sets.loc_techs_supply_plus
        and constraint_exists(model_run, i, 'constraints.resource_area_per_energy_cap')
        """
        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'resource_area_per_energy_capacity_constraint')

        m = build_model(
            {'techs.test_supply_plus.constraints.resource_area_max': 10},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'resource_area_per_energy_capacity_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.resource_area_per_energy_cap': 10},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'resource_area_per_energy_capacity_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints': {'resource_area_per_energy_cap': 10,
                                                    'resource_area_max': 10}},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'resource_area_per_energy_capacity_constraint')

    def test_locs_resource_area_capacity_per_loc_constraint(self):
        """
        i for i in sets.locs
        if model_run.locations[i].get_key('available_area', None) is not None
        """
        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'resource_area_capacity_per_loc_constraint')

        m = build_model({'locations.0.available_area': 1}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'resource_area_capacity_per_loc_constraint')

        m = build_model(
            {'locations.0.available_area': 1,
             'techs.test_supply_plus.constraints.resource_area_max': 10},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'resource_area_capacity_per_loc_constraint')

    def test_loc_techs_energy_capacity_constraint(self):
        """
        i for i in sets.loc_techs
        if i not in sets.loc_techs_milp + sets.loc_techs_purchase
        """
        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_constraint')

        m2 = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_scale': 5},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m2.run(build_only=True)
        assert (
            m2._backend_model.energy_capacity_constraint['0::test_supply_elec'].upper() ==
            m._backend_model.energy_capacity_constraint['0::test_supply_elec'].upper() * 5
        )

        m = build_model({}, 'supply_milp,two_hours,investment_costs')  # demand still is in loc_techs
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_constraint')

        # Check that setting `_equals` to infinity is caught:
        override = {'locations.0.techs.test_supply_elec.constraints.energy_cap_equals': np.inf}
        with pytest.raises(exceptions.ModelError) as error:
            m = build_model(override, 'simple_supply,two_hours,investment_costs')
            m.run(build_only=True)

        assert check_error_or_warning(
            error,
            'Cannot use inf for energy_cap_equals for loc:tech `0::test_supply_elec`'
        )

    def test_techs_energy_capacity_systemwide_constraint(self):
        """
        i for i in sets.techs
        if model_run.get_key('techs.{}.constraints.energy_cap_max_systemwide'.format(i), None)
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'energy_capacity_systemwide_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_max_systemwide': 20},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_systemwide_constraint')
        assert (
            'test_supply_elec' in
            m._backend_model.energy_capacity_systemwide_constraint.keys()
        )

        # setting the constraint to infinity leads to Pyomo creating NoConstraint
        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_max_systemwide': np.inf},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_systemwide_constraint')
        assert (
            'test_supply_elec' not in
            m._backend_model.energy_capacity_systemwide_constraint.keys()
        )

        # Check that setting `_equals` to infinity is caught:
        with pytest.raises(exceptions.ModelError) as error:
            m = build_model(
                {'techs.test_supply_elec.constraints.energy_cap_equals_systemwide': np.inf},
                'simple_supply,two_hours,investment_costs'
            )
            m.run(build_only=True)

        assert check_error_or_warning(
            error,
            'Cannot use inf for energy_cap_equals_systemwide for tech `test_supply_elec`'
        )

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_equals_systemwide': 20},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_systemwide_constraint')


class TestDispatchConstraints:
    # dispatch.py
    def test_loc_tech_carriers_carrier_production_max_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_production_max_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_max_constraint')

    def test_loc_tech_carriers_carrier_production_min_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min_use': 0.1},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_production_min_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min_use': 0.1},
            'supply_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_constraint')

    def test_loc_tech_carriers_carrier_consumption_max_constraint(self):
        """
        i for i in sets.loc_tech_carriers_con
        if i.rsplit('::', 1)[0] in sets.loc_techs_demand +
            sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
        """

        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_consumption_max_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_consumption_max_constraint')

    def test_loc_techs_resource_max_constraint(self):
        """
        sets.loc_techs_finite_resource_supply_plus,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'resource_max_constraint')

        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'resource_max_constraint')

        m = build_model(
            {'techs.test_supply_plus.constraints.resource': np.inf},
            'simple_supply_and_supply_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'resource_max_constraint')

    def test_loc_techs_storage_max_constraint(self):
        """
        sets.loc_techs_store
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_max_constraint')

        m = build_model({}, 'simple_supply_and_supply_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'storage_max_constraint')

        m = build_model({}, 'simple_storage,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'storage_max_constraint')

    def test_loc_tech_carriers_ramping_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i.rsplit('::', 1)[0] in sets.loc_techs_ramping
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'ramping_up_constraint')
        assert not hasattr(m._backend_model, 'ramping_down_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_ramping': 0.1},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'ramping_up_constraint')
        assert hasattr(m._backend_model, 'ramping_down_constraint')

        m = build_model(
            {'techs.test_conversion.constraints.energy_ramping': 0.1},
            'simple_conversion,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'ramping_up_constraint')
        assert hasattr(m._backend_model, 'ramping_down_constraint')


class TestMILPConstraints:
    # milp.py
    def test_loc_techs_unit_commitment_constraint(self):
        """
        sets.loc_techs_milp,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'unit_commitment_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'unit_commitment_constraint')

        m = build_model({}, 'supply_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'unit_commitment_constraint')

    def test_loc_techs_unit_capacity_constraint(self):
        """
        sets.loc_techs_milp,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'unit_capacity_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'unit_capacity_constraint')

        m = build_model({}, 'supply_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'unit_capacity_constraint')

    def test_loc_tech_carriers_carrier_production_max_milp_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_max_milp_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_production_max_milp_constraint')

        m = build_model({}, 'supply_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_max_milp_constraint')

        m = build_model({}, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_max_milp_constraint')

    def test_loc_techs_carrier_production_max_conversion_plus_milp_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if i in sets.loc_techs_milp
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model,
            'carrier_production_max_conversion_plus_milp_constraint'
        )

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model,
            'carrier_production_max_conversion_plus_milp_constraint'
        )

        m = build_model({}, 'supply_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model,
            'carrier_production_max_conversion_plus_milp_constraint'
        )

        m = build_model({}, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(
            m._backend_model,
            'carrier_production_max_conversion_plus_milp_constraint'
        )

        m = build_model({}, 'conversion_plus_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model,
            'carrier_production_max_conversion_plus_milp_constraint'
        )

    def test_loc_tech_carriers_carrier_production_min_milp_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(model_run, i.rsplit('::', 1)[0], 'constraints.energy_cap_min_use')
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
        """
        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min_use': 0.1},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_milp_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min_use': 0.1},
            'supply_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_production_min_milp_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min_use': 0.1},
            'supply_purchase,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_milp_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min_use': 0.1},
            'conversion_plus_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_milp_constraint')

        m = build_model(
            {'techs.test_conversion_plus.constraints.energy_cap_min_use': 0.1},
            'conversion_plus_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_milp_constraint')

    def test_loc_techs_carrier_production_min_conversion_plus_milp_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i in sets.loc_techs_milp
        """
        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min_use': 0.1},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model,
            'carrier_production_min_conversion_plus_milp_constraint'
        )

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min_use': 0.1},
            'supply_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model,
            'carrier_production_min_conversion_plus_milp_constraint'
        )

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min_use': 0.1},
            'conversion_plus_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model,
            'carrier_production_min_conversion_plus_milp_constraint'
        )

        m = build_model(
            {'techs.test_conversion_plus.constraints.energy_cap_min_use': 0.1},
            'conversion_plus_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(
            m._backend_model,
            'carrier_production_min_conversion_plus_milp_constraint'
        )

        m = build_model(
            {'techs.test_conversion_plus.constraints.energy_cap_min_use': 0.1},
            'conversion_plus_purchase,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model,
            'carrier_production_min_conversion_plus_milp_constraint'
        )

    def test_loc_tech_carriers_carrier_consumption_max_milp_constraint(self):
        """
        i for i in sets.loc_tech_carriers_con
        if i.rsplit('::', 1)[0] in sets.loc_techs_demand +
            sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_consumption_max_milp_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_consumption_max_milp_constraint')

        m = build_model({}, 'storage_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_consumption_max_milp_constraint')

        m = build_model({}, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_consumption_max_milp_constraint')

    def test_loc_techs_energy_capacity_units_constraint(self):
        """
        i for i in sets.loc_techs_milp
        if constraint_exists(model_run, i, 'constraints.energy_cap_per_unit')
        is not None
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'energy_capacity_units_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_units_constraint')

        m = build_model({}, 'storage_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_units_constraint')

        m = build_model({}, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_units_constraint')

    def test_loc_techs_storage_capacity_units_constraint(self):
        """
        i for i in sets.loc_techs_milp if i in sets.loc_techs_store
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_units_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_units_constraint')

        m = build_model({}, 'storage_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'storage_capacity_units_constraint')

        m = build_model({}, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_units_constraint')

        m = build_model({}, 'supply_and_supply_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'storage_capacity_units_constraint')

    def test_loc_techs_energy_capacity_max_purchase_constraint(self):
        """
        i for i in sets.loc_techs_purchase
        if (constraint_exists(model_run, i, 'constraints.energy_cap_equals') is not None
            or constraint_exists(model_run, i, 'constraints.energy_cap_max') is not None)
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'energy_capacity_max_purchase_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'energy_capacity_max_purchase_constraint')

        m = build_model({}, 'supply_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_max_purchase_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints': {'energy_cap_max': None, 'energy_cap_equals': 15}},
            'supply_purchase,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_max_purchase_constraint')

    def test_loc_techs_energy_capacity_min_purchase_constraint(self):
        """
        i for i in sets.loc_techs_purchase
        if (not constraint_exists(model_run, i, 'constraints.energy_cap_equals')
            and constraint_exists(model_run, i, 'constraints.energy_cap_min'))
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'energy_capacity_min_purchase_constraint')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'energy_capacity_min_purchase_constraint')

        m = build_model({}, 'supply_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'energy_capacity_min_purchase_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints': {'energy_cap_max': None, 'energy_cap_equals': 15}},
            'supply_purchase,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'energy_capacity_min_purchase_constraint')

        m = build_model(
            {'techs.test_supply_elec.constraints.energy_cap_min': 10},
            'supply_purchase,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'energy_capacity_min_purchase_constraint')

    def test_loc_techs_storage_capacity_max_purchase_constraint(self):
        """
        i for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        """
        m = build_model({}, 'simple_storage,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_max_purchase_constraint')

        m = build_model({}, 'storage_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_max_purchase_constraint')

        m = build_model({}, 'storage_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'storage_capacity_max_purchase_constraint')

        m = build_model({}, 'supply_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_max_purchase_constraint')

    def test_loc_techs_storage_capacity_min_purchase_constraint(self):
        """
        i for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        if (not constraint_exists(model_run, i, 'constraints.storage_cap_equals')
            and (constraint_exists(model_run, i, 'constraints.storage_cap_min')
                or constraint_exists(model_run, i, 'constraints.energy_cap_min')))
        """
        m = build_model(
            {'techs.test_storage.constraints.storage_cap_min': 10},
            'simple_storage,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_min_purchase_constraint')

        m = build_model(
            {'techs.test_storage.constraints.storage_cap_min': 10},
            'storage_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_min_purchase_constraint')

        m = build_model({}, 'storage_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_min_purchase_constraint')

        m = build_model(
            {'techs.test_storage.constraints.storage_cap_min': 10},
            'storage_purchase,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'storage_capacity_min_purchase_constraint')

        m = build_model(
            {'techs.test_storage.constraints': {'storage_cap_equals': 10, 'storage_cap_min': 10}},
            'storage_purchase,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'storage_capacity_min_purchase_constraint')

    def test_loc_techs_update_costs_investment_units_constraint(self):
        """
        i for i in sets.loc_techs_milp
        if i in sets.loc_techs_investment_cost and
        any(constraint_exists(model_run, i, 'costs.{}.purchase'.format(j))
               for j in model_run.sets.costs)
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not check_variable_exists(m._backend_model, 'cost_investment_constraint', 'purchased')
        assert not check_variable_exists(m._backend_model, 'cost_investment_constraint', 'units')

        m = build_model({}, 'supply_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not check_variable_exists(m._backend_model, 'cost_investment_constraint', 'purchased')
        assert not check_variable_exists(m._backend_model, 'cost_investment_constraint', 'units')

        m = build_model(
            {'techs.test_supply_elec.costs.monetary.purchase': 1},
            'supply_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not check_variable_exists(m._backend_model, 'cost_investment_constraint', 'purchased')
        assert check_variable_exists(m._backend_model, 'cost_investment_constraint', 'units')

    def test_loc_techs_update_costs_investment_purchase_constraint(self):
        """
        sets.loc_techs_purchase,
        """

        m = build_model({}, 'supply_purchase,two_hours,investment_costs')
        m.run(build_only=True)
        assert check_variable_exists(m._backend_model, 'cost_investment_constraint', 'purchased')
        assert not check_variable_exists(m._backend_model, 'cost_investment_constraint', 'units')

    def test_techs_unit_capacity_systemwide_constraint(self):
        """
        sets.techs if unit_cap_max_systemwide or unit_cap_equals_systemwide
        """

        override_max = {
            'links.0,1.exists': True,
            'techs.test_conversion_plus.constraints.units_max_systemwide': 2,
            'locations.1.techs.test_conversion_plus.constraints': {
                'units_max': 2,
                'energy_cap_per_unit': 5
            }
        }
        override_equals = {
            'links.0,1.exists': True,
            'techs.test_conversion_plus.constraints.units_equals_systemwide': 1,
            'locations.1.techs.test_conversion_plus.costs.monetary.purchase': 1
        }
        override_equals_inf = {
            'links.0,1.exists': True,
            'techs.test_conversion_plus.constraints.units_equals_systemwide': np.inf,
            'locations.1.techs.test_conversion_plus.costs.monetary.purchase': 1
        }

        m = build_model(override_max, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'unit_capacity_systemwide_constraint')
        assert m._backend_model.unit_capacity_systemwide_constraint['test_conversion_plus'].upper() == 2

        m = build_model(override_equals, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'unit_capacity_systemwide_constraint')
        assert m._backend_model.unit_capacity_systemwide_constraint['test_conversion_plus'].lower() == 1
        assert m._backend_model.unit_capacity_systemwide_constraint['test_conversion_plus'].upper() == 1

        with pytest.raises(ValueError) as error:
            m = build_model(override_equals_inf, 'conversion_plus_milp,two_hours,investment_costs')
            m.run(build_only=True)
        assert check_error_or_warning(error, 'Cannot use inf for energy_cap_equals_systemwide')


class TestConversionConstraints:

    # conversion.py
    def test_loc_techs_balance_conversion_constraint(self):
        """
        sets.loc_techs_conversion,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_constraint')

        m = build_model({}, 'simple_conversion,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_constraint')

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_constraint')

    def test_loc_techs_cost_var_conversion_constraint(self):
        """
        sets.loc_techs_om_cost_conversion,
        """
        m = build_model(
            {'techs.test_supply_elec.costs.monetary.om_prod': 0.1},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'cost_var_conversion_constraint')

        m = build_model({}, 'simple_conversion,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'cost_var_conversion_constraint')

        m = build_model(
            {'techs.test_conversion.costs.monetary.om_prod': 0.1},
            'simple_conversion,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_conversion_constraint')

        assert check_variable_exists(
            m._backend_model, 'cost_var_conversion_constraint', 'carrier_prod'
        )
        assert not check_variable_exists(
            m._backend_model, 'cost_var_conversion_constraint', 'carrier_con'
        )

        m = build_model(
            {'techs.test_conversion.costs.monetary.om_con': 0.1},
            'simple_conversion,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_conversion_constraint')
        assert check_variable_exists(
            m._backend_model, 'cost_var_conversion_constraint', 'carrier_con'
        )
        assert not check_variable_exists(
            m._backend_model, 'cost_var_conversion_constraint', 'carrier_prod'
        )

class TestConversionPlusConstraints:
    # conversion_plus.py
    def test_loc_techs_balance_conversion_plus_primary_constraint(self):
        """
        sets.loc_techs_conversion_plus,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_plus_primary_constraint')

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_primary_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials.carrier_out': ['electricity', 'heat']},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_primary_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in': ['coal', 'gas'], 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_primary_constraint')

    def test_loc_techs_carrier_production_max_conversion_plus_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if i not in sets.loc_techs_milp
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_production_max_conversion_plus_constraint')

        m = build_model({}, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_max_conversion_plus_constraint')

    def test_loc_techs_carrier_production_min_conversion_plus_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i not in sets.loc_techs_milp
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_conversion_plus_constraint')

        m = build_model({}, 'conversion_plus_milp,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_conversion_plus_constraint')

        m = build_model(
            {'techs.test_conversion_plus.constraints.energy_cap_min_use': 0.1},
            'conversion_plus_milp,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'carrier_production_min_conversion_plus_constraint')

        m = build_model(
            {'techs.test_conversion_plus.constraints.energy_cap_min_use': 0.1},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'carrier_production_min_conversion_plus_constraint')

    def test_loc_techs_cost_var_conversion_plus_constraint(self):
        """
        sets.loc_techs_om_cost_conversion_plus,
        """
        # no conversion_plus = no constraint
        m = build_model(
            {'techs.test_supply_elec.costs.monetary.om_prod': 0.1},
            'simple_supply,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')

        # no conversion_plus = no constraint
        m = build_model(
            {'techs.test_conversion.costs.monetary.om_prod': 0.1},
            'simple_conversion,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')

        # no variable costs for conversion_plus = no constraint
        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')

        # om_prod creates constraint and populates it with carrier_prod driven cost
        m = build_model(
            {'techs.test_conversion_plus.costs.monetary.om_prod': 0.1},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')
        assert check_variable_exists(
            m._backend_model, 'cost_var_conversion_plus_constraint', 'carrier_prod'
        )
        assert not check_variable_exists(
            m._backend_model, 'cost_var_conversion_plus_constraint', 'carrier_con'
        )

        # om_con creates constraint and populates it with carrier_con driven cost
        m = build_model(
            {'techs.test_conversion_plus.costs.monetary.om_con': 0.1},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'cost_var_conversion_plus_constraint')
        assert check_variable_exists(
            m._backend_model, 'cost_var_conversion_plus_constraint', 'carrier_con'
        )
        assert not check_variable_exists(
            m._backend_model, 'cost_var_conversion_plus_constraint', 'carrier_prod'
        )

    def test_loc_techs_balance_conversion_plus_in_2_constraint(self):
        """
        sets.loc_techs_in_2,
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_plus_in_2_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in_2': 'coal', 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_in_2_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in_2': ['coal', 'heat'], 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_in_2_constraint')

    def test_loc_techs_balance_conversion_plus_in_3_constraint(self):
        """
        sets.loc_techs_in_3,
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_plus_in_3_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in_3': 'coal', 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_in_3_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials': {
                'carrier_in_3': ['coal', 'heat'], 'primary_carrier_in': 'gas'
            }},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_in_3_constraint')

    def test_loc_techs_balance_conversion_plus_out_2_constraint(self):
        """
        sets.loc_techs_out_2,
        """

        m = build_model(
            {'techs.test_conversion_plus.essentials.carrier_out_2': ['coal', 'heat']},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_out_2_constraint')

    def test_loc_techs_balance_conversion_plus_out_3_constraint(self):
        """
        sets.loc_techs_out_3,
        """

        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'balance_conversion_plus_out_3_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials.carrier_out_3': 'coal'},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_out_3_constraint')

        m = build_model(
            {'techs.test_conversion_plus.essentials.carrier_out_3': ['coal', 'heat']},
            'simple_conversion_plus,two_hours,investment_costs'
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'balance_conversion_plus_out_3_constraint')


class TestNetworkConstraints:
    # network.py
    def test_loc_techs_symmetric_transmission_constraint(self):
        """
        sets.loc_techs_transmission,
        """
        m = build_model({}, 'simple_supply,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'symmetric_transmission_constraint')


        m = build_model({}, 'simple_conversion_plus,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'symmetric_transmission_constraint')


class TestPolicyConstraints:
    # policy.py
    def test_techlists_group_share_energy_cap_min_constraint(self):
        """
        i for i in sets.techlists
        if 'energy_cap_min' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        """
        m = build_model({}, 'simple_supply,group_share_energy_cap_min,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'group_share_energy_cap_min_constraint')
        assert not hasattr(m._backend_model, 'group_share_energy_cap_max_constraint')
        assert not hasattr(m._backend_model, 'group_share_energy_cap_equals_constraint')

    def test_techlists_group_share_energy_cap_max_constraint(self):
        """
        i for i in sets.techlists
        if 'energy_cap_max' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        """
        m = build_model({}, 'simple_supply,group_share_energy_cap_max,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'group_share_energy_cap_min_constraint')
        assert hasattr(m._backend_model, 'group_share_energy_cap_max_constraint')
        assert not hasattr(m._backend_model, 'group_share_energy_cap_equals_constraint')

    def test_techlists_group_share_energy_cap_equals_constraint(self):
        """
        i for i in sets.techlists
        if 'energy_cap_equals' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        """
        m = build_model({}, 'simple_supply,group_share_energy_cap_equals,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'group_share_energy_cap_min_constraint')
        assert not hasattr(m._backend_model, 'group_share_energy_cap_max_constraint')
        assert hasattr(m._backend_model, 'group_share_energy_cap_equals_constraint')

    def test_techlists_carrier_group_share_carrier_prod_min_constraint(self):
        """
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_min' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_min'.format(i), {}).keys()
        """

        m = build_model({}, 'conversion_and_conversion_plus,group_share_carrier_prod_min,two_hours,investment_costs')
        m.run(build_only=True)
        assert hasattr(m._backend_model, 'group_share_carrier_prod_min_constraint')
        assert not hasattr(m._backend_model, 'group_share_carrier_prod_max_constraint')
        assert not hasattr(m._backend_model, 'group_share_carrier_prod_equals_constraint')

    def test_techlists_carrier_group_share_carrier_prod_max_constraint(self):
        """
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_max' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_max'.format(i), {}).keys()
        """

        m = build_model({}, 'conversion_and_conversion_plus,group_share_carrier_prod_max,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'group_share_carrier_prod_min_constraint')
        assert hasattr(m._backend_model, 'group_share_carrier_prod_max_constraint')
        assert not hasattr(m._backend_model, 'group_share_carrier_prod_equals_constraint')

    def test_techlists_carrier_group_share_carrier_prod_equals_constraint(self):
        """
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_equals' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_equals'.format(i), {}).keys()
        """

        m = build_model({}, 'conversion_and_conversion_plus,group_share_carrier_prod_equals,two_hours,investment_costs')
        m.run(build_only=True)
        assert not hasattr(m._backend_model, 'group_share_carrier_prod_min_constraint')
        assert not hasattr(m._backend_model, 'group_share_carrier_prod_max_constraint')
        assert hasattr(m._backend_model, 'group_share_carrier_prod_equals_constraint')


# clustering constraints
class TestClusteringConstraints:

    def constraints(self):
        return ['balance_storage_inter_cluster_constraint',
                'storage_intra_max_constraint', 'storage_intra_min_constraint',
                'storage_inter_max_constraint', 'storage_inter_min_constraint']

    def decision_variables(self):
        return ['storage_inter_cluster',
                'storage_intra_cluster_max', 'storage_intra_cluster_min']

    def cluster_model(self, how='mean', storage_inter_cluster=True,
                      cyclic=False, storage_initial=False):
        override = {
            'model.subset_time': ['2005-01-01', '2005-01-04'],
            'model.time': {
                'function': 'apply_clustering',
                'function_options': {
                    'clustering_func': 'file=cluster_days.csv:0', 'how': how,
                    'storage_inter_cluster': storage_inter_cluster
                }
            },
            'run.cyclic_storage': cyclic
        }
        if storage_initial:
            override.update({'techs.test_storage.constraints.storage_initial': 0})
        return build_model(override, 'simple_storage,investment_costs')

    def test_cluster_storage_constraints(self):
        m = self.cluster_model()
        m.run(build_only=True)

        for variable in self.decision_variables():
            assert hasattr(m._backend_model, variable)

        for constraint in self.constraints():
            assert hasattr(m._backend_model, constraint)

        assert not hasattr(m._backend_model, 'storage_max_constraint')
        assert not hasattr(m._backend_model, 'storage_initial_constraint')

    def test_cluster_cyclic_storage_constraints(self):
        m = self.cluster_model(cyclic=True)
        m.run(build_only=True)

        for variable in self.decision_variables():
            assert hasattr(m._backend_model, variable)

        for constraint in self.constraints():
            assert hasattr(m._backend_model, constraint)

        assert not hasattr(m._backend_model, 'storage_max_constraint')
        assert not hasattr(m._backend_model, 'storage_initial_constraint')

    def test_no_cluster_storage_constraints(self):
        m = self.cluster_model(storage_inter_cluster=False)
        m.run(build_only=True)

        for variable in self.decision_variables():
            assert not hasattr(m._backend_model, variable)

        for constraint in self.constraints():
            assert not hasattr(m._backend_model, constraint)

        assert hasattr(m._backend_model, 'storage_max_constraint')



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

funcs.py
~~~~~~~~

Functions to process time series data.

"""

import numpy as np
import pandas as pd
import xarray as xr

from calliope import exceptions
from calliope.core.util.dataset import get_loc_techs
from calliope.core.time import clustering
from calliope.core.util.logging import logger
from calliope.core.preprocess.lookup import lookup_clusters



def get_daily_timesteps(data, check_uniformity=False):
    daily_timesteps = [
        data.timestep_resolution.loc[i].values
        for i in np.unique(data.timesteps.to_index().strftime('%Y-%m-%d'))
    ]

    if check_uniformity:
        if not np.all(daily_timesteps == daily_timesteps[0]):
            raise exceptions.ModelError(
                'For clustering, timestep resolution must be uniform.'
            )

    return daily_timesteps[0]


def normalized_copy(data):
    """
    Normalize timeseries data, using the maximum across all regions and timesteps.

    Parameters
    ----------
    data : xarray Dataset
        Dataset with all non-time dependent variables removed

    Returns
    -------
    ds : xarray Dataset
        Copy of `data`, with the absolute taken and normalized to 0-1

    """
    ds = data.copy(deep=True)  # Work off a copy

    for var in ds.data_vars:
        # Each DataArray is indexed over a different subset of loc_techs,
        # so we find it in the list of dimensions
        loc_tech_dim = [i for i in ds[var].dims if 'loc_techs' in i][0]

        # For each technology, get the loc_techs which are relevant
        loc_tech_subsets = [
            get_loc_techs(ds[loc_tech_dim].values, tech)
            for tech in set(i.split('::')[1] for i in ds[loc_tech_dim].values)
        ]
        # remove empty lists within the _techs list
        loc_tech_subsets = [i for i in loc_tech_subsets if i]

        # For each technology, divide all values by the maximum absolute value
        for loc_tech in loc_tech_subsets:
            ds[var].loc[{loc_tech_dim: loc_tech}] = abs(
                ds[var].loc[{loc_tech_dim: loc_tech}] /
                abs(ds[var].loc[{loc_tech_dim: loc_tech}]).max()
            )
    return ds


def _copy_non_t_vars(data0, data1):
    """Copies non-t-indexed variables from data0 into data1, then
    returns data1"""
    non_t_vars = [v for v in data0.data_vars
                  if 'timesteps' not in data0[v].dims]
    # Manually copy over variables not in `t`. If we don't do this,
    # these vars get polluted with a superfluous `t` dimension
    for v in non_t_vars:
        data1[v] = data0[v]
    return data1


def _combine_datasets(data0, data1):
    """Concatenates data0 and data1 along the time dimension"""
    data_new = xr.concat([data0, data1], dim='timesteps')
    # Ensure time dimension is ordered
    data_new = data_new.loc[
        {'timesteps': data_new.timesteps.to_index().sort_values()}
    ]

    return data_new


def apply_clustering(data, timesteps, clustering_func, how, normalize=True,
                     scale_clusters='mean', storage_inter_cluster=True,
                     model_run=None, **kwargs):
    """
    Apply the given clustering function to the given data.

    Parameters
    ----------
    data : xarray.Dataset
    timesteps : pandas.DatetimeIndex or list of timesteps or None
    clustering_func : str
        Name of clustering function. Can be `file=....csv:column_name`
        if loading custom clustering. Custom clustering index = timeseries days.
        If no column_name, the CSV file must have only one column of data.
    how : str
        How to map clusters to data. 'mean' or 'closest'.
    normalize : bool, optional
        If True (default), data is normalized before clustering is applied,
        using :func:`~calliope.core.time.funcs.normalized_copy`.
    scale_clusters : str or None, default = 'mean'
        Scale the results of clustering such that the clusters match the metric
        given by scale_clusters. For example, 'mean' scales along each loc_tech
        and variable to match inputs and outputs. Other options for matching
        include 'sum', 'max', and 'min'. If None, no scaling occurs.
    **kwargs : optional
        Arguments passed to clustering_func.

    Returns
    -------
    data_new_scaled : xarray.Dataset

    """

    assert how in ['mean', 'closest']

    daily_timesteps = get_daily_timesteps(data, check_uniformity=True)
    timesteps_per_day = len(daily_timesteps)

    # Save all coordinates, to ensure they can be added back in after clustering
    data_coords = data.copy().coords
    del data_coords['timesteps']
    # Only apply clustering function on subset of masked timesteps
    if timesteps is None:
        data_to_cluster = data
    else:
        data_to_cluster = data.loc[{'timesteps': timesteps}]

    # remove all variables that are not indexed over time
    data_to_cluster = data_to_cluster.drop([
        i for i in data.variables
        if 'timesteps' not in data[i].dims or 'timestep_' in i
    ])

    for dim in data_to_cluster.dims:
        data_to_cluster[dim] = data[dim]

    if normalize:
        data_normalized = normalized_copy(data_to_cluster)
    else:
        data_normalized = data_to_cluster

    if 'file=' in clustering_func:
        file = clustering_func.split('=')[1]
        if ':' in file:
            file, column = file.rsplit(':', 1)
        else:
            column = None

        df = model_run.timeseries_data[file]
        if isinstance(df, pd.Series) and column is not None:
            raise exceptions.ModelWarning(
                '{} given as time clustering column, but only one column to '
                'choose from in {}.'.format(column, file)
            )
            clusters = df.resample('1D').mean()
        elif isinstance(df, pd.DataFrame) and column is None:
            raise exceptions.ModelError(
                'No time clustering column given, but multiple columns found in '
                '{0}. Choose one column and add it to {1} as {1}:name_of_column.'
                .format(file, clustering_func)
            )
        elif isinstance(df, pd.DataFrame) and column not in df.columns:
            raise KeyError(
                'time clustering column {} not found in {}.'.format(column, file)
            )
        elif isinstance(df, pd.DataFrame):
            clusters = df.loc[:, column].groupby(pd.Grouper(freq='1D')).unique()

        # Check there weren't instances of more than one cluster assigned to a day
        # or days with no information assigned
        if any([len(i) == 0 for i in clusters.values]):
            raise exceptions.ModelError(
                'Missing cluster days in `{}:{}`.'.format(file, column)
            )
        elif any([len(i) > 1 for i in clusters.values]):
            raise exceptions.ModelError(
                'More than one cluster value assigned to a day in `{}:{}`. '
                'Unique clusters per day: {}'.format(file, column, clusters)
            )
        else:
            clusters.loc[:] = [i[0] for i in clusters.values]

    else:
        result = clustering.get_clusters(
            data_normalized, clustering_func, timesteps_per_day=timesteps_per_day,
            **kwargs
        )
        clusters = result[0]  # Ignore other stuff returned

    data_new = clustering.map_clusters_to_data(
        data_to_cluster, clusters,
        how=how, daily_timesteps=daily_timesteps,
        storage_inter_cluster=storage_inter_cluster
    )

    if timesteps is None:
        data_new = _copy_non_t_vars(data, data_new)
    else:
        # Drop timesteps from old data
        data_new = _copy_non_t_vars(data, data_new)
        data_new = _combine_datasets(data.drop(timesteps, dim='timesteps'), data_new)
        data_new = _copy_non_t_vars(data, data_new)

    # It's now safe to add the original coordinates back in (preserving all the
    # loc_tech sets that aren't used to index a variable in the DataArray)
    data_new.update(data_coords)

    # Scale the new/combined data so that the mean for each (loc_tech, variable)
    # combination matches that from the original data
    data_new_scaled = data_new.copy(deep=True)
    if scale_clusters:
        data_vars_in_t = [
            v for v in data_new.data_vars
            if 'timesteps' in data_new[v].dims and
            'timestep_' not in v and v != 'clusters'
        ]
        for var in data_vars_in_t:
            scale = (
                getattr(data[var], scale_clusters)(dim='timesteps') /
                getattr(data_new[var], scale_clusters)(dim='timesteps')
            )
            data_new_scaled[var] = data_new[var] * scale.fillna(0)

    lookup_clusters(data_new_scaled)

    return data_new_scaled


def resample(data, timesteps, resolution):
    """
    Function to resample timeseries data from the input resolution (e.g. 1H), to
    the given resolution (e.g. 2H)

    Parameters
    ----------
    data : xarray.Dataset
        calliope model data, containing only timeseries data variables
    timesteps : str or list; optional
        If given, apply resampling to a subset of the timeseries data
    resolution : str
        time resolution of the output data, given in Pandas time frequency format.
        E.g. 1H = 1 hour, 1W = 1 week, 1M = 1 month, 1T = 1 minute. Multiples allowed.

    """
    data_new = data.copy(deep=True)
    if timesteps is not None:
        data_new = data_new.loc[{'timesteps': timesteps}]

    # First create a new resampled dataset of the correct size by
    # using first-resample, which should be a quick way to achieve this
    data_rs = data_new.resample(resolution, dim='timesteps', how='first')

    timestep_vars = [v for v in data_new.data_vars
                     if 'timesteps' in data_new[v].dims]

    # Resampling adds spurious `time` dimension to non-time vars, correct that
    for v in data_rs.data_vars:
        if v not in timestep_vars:
            data_rs[v] = data[v]

    for var in timestep_vars:
        if var in ['timestep_resolution', 'resource']:
            data_rs[var] = data_new[var].resample(
                resolution, dim='timesteps', how='sum'
            )
        else:
            try:
                data_rs[var] = data_new[var].resample(
                    resolution, dim='timesteps', how='mean'
                )
            except TypeError:
                # If the var has a datatype of strings, it can't be resampled
                logger.error('Dropping {} because it has a {} data type when '
                              'integer or float is expected for timeseries '
                              'resampling.'.format(var, data_rs[var].dtype))
                data_rs = data_rs.drop(var)

    # Get rid of the filled-in NaN timestamps
    data_rs = data_rs.dropna(dim='timesteps', how='all')

    # repopulate the attribute dictionary, as it will have been lost along the way
    data_rs.attrs.update(data.attrs)
    data_rs.attrs['allow_operate_mode'] = 1  # Resampling still permits operational mode

    if timesteps is not None:
        # Combine leftover parts of passed in data with new data
        data_rs = _copy_non_t_vars(data, data_rs)
        data_rs = _combine_datasets(data.drop(timesteps, dim='timesteps'), data_rs)
        data_rs = _copy_non_t_vars(data, data_rs)
        # Having timesteps with different lengths does not permit operational mode
        data_rs.attrs['allow_operate_mode'] = 0

    return data_rs


def drop(data, timesteps):
    """
    Drop timesteps from data, adjusting the timestep weight of remaining
    timesteps accordingly. Returns updated dataset.

    Parameters
    ----------
    data : xarray.Dataset
        Calliope model data.
    timesteps : str or list or other iterable
        Pandas-compatible timestep strings.

    """
    # Turn timesteps into a pandas datetime index for subsetting, which also
    # checks whether they are actually valid
    try:
        timesteps_pd = pd.to_datetime(timesteps)
    except Exception as e:
        raise exceptions.ModelError(
            'Invalid timesteps: {}'.format(timesteps)
        )

    # 'Distribute weight' of the dropped timesteps onto the remaining ones
    dropped_weight = data.timestep_weights.loc[{'timesteps': timesteps_pd}].sum()

    data = data.drop(timesteps_pd, dim='timesteps')

    data['timestep_weights'] = data['timestep_weights'] + (dropped_weight / len(data['timestep_weights']))

    return data



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

import pyomo.core as po  # pylint: disable=import-error


def initialize_decision_variables(backend_model):
    """
    Defines decision variables.

    ==================== ========================================
    Variable             Dimensions
    ==================== ========================================
    energy_cap           loc_techs
    carrier_prod         loc_tech_carriers_prod, timesteps
    carrier_con          loc_tech_carriers_con, timesteps
    cost                 costs, loc_techs_cost
    resource_area        loc_techs_area,
    storage_cap          loc_techs_store
    storage              loc_techs_store, timesteps
    resource_con         loc_techs_supply_plus, timesteps
    resource_cap         loc_techs_supply_plus
    carrier_export       loc_tech_carriers_export, timesteps
    cost_var             costs, loc_techs_om_cost, timesteps
    cost_investment      costs, loc_techs_investment_cost
    purchased            loc_techs_purchase
    units                loc_techs_milp
    operating\_units     loc_techs_milp, timesteps
    unmet\_demand        loc_carriers, timesteps
    ==================== ========================================

    """
    model_data_dict = backend_model.__calliope_model_data__

    ##
    # Variables which are always assigned
    ##
    if backend_model.mode != 'operate':
        backend_model.energy_cap = po.Var(backend_model.loc_techs, within=po.NonNegativeReals)
    backend_model.carrier_prod = po.Var(backend_model.loc_tech_carriers_prod, backend_model.timesteps, within=po.NonNegativeReals)
    backend_model.carrier_con = po.Var(backend_model.loc_tech_carriers_con, backend_model.timesteps, within=po.NegativeReals)
    backend_model.cost = po.Var(backend_model.costs, backend_model.loc_techs_cost, within=po.Reals)

    ##
    # Conditionally assigned variables
    ##

    if 'loc_techs_area' in model_data_dict['sets'] and backend_model.mode != 'operate':
        backend_model.resource_area = po.Var(backend_model.loc_techs_area, within=po.NonNegativeReals)

    if 'loc_techs_store' in model_data_dict['sets']:
        if backend_model.mode != 'operate':
            backend_model.storage_cap = po.Var(backend_model.loc_techs_store, within=po.NonNegativeReals)
        if hasattr(backend_model, 'clusters') and hasattr(backend_model, 'datesteps'):
            backend_model.storage_inter_cluster = po.Var(backend_model.loc_techs_store, backend_model.datesteps, within=po.NonNegativeReals)
            backend_model.storage_intra_cluster_max = po.Var(backend_model.loc_techs_store, backend_model.clusters, within=po.Reals)
            backend_model.storage_intra_cluster_min = po.Var(backend_model.loc_techs_store, backend_model.clusters, within=po.Reals)
            storage_within = po.Reals
        else:
            storage_within = po.NonNegativeReals
        backend_model.storage = po.Var(backend_model.loc_techs_store, backend_model.timesteps, within=storage_within)

    if 'loc_techs_supply_plus' in model_data_dict['sets']:
        backend_model.resource_con = po.Var(backend_model.loc_techs_supply_plus, backend_model.timesteps, within=po.NonNegativeReals)
        if backend_model.mode != 'operate':
            backend_model.resource_cap = po.Var(backend_model.loc_techs_supply_plus, within=po.NonNegativeReals)

    if 'loc_techs_export' in model_data_dict['sets']:
        backend_model.carrier_export = po.Var(backend_model.loc_tech_carriers_export, backend_model.timesteps, within=po.NonNegativeReals)

    if 'loc_techs_om_cost' in model_data_dict['sets']:
        backend_model.cost_var = po.Var(backend_model.costs, backend_model.loc_techs_om_cost, backend_model.timesteps, within=po.Reals)

    if 'loc_techs_investment_cost' in model_data_dict['sets'] and backend_model.mode != 'operate':
        backend_model.cost_investment = po.Var(backend_model.costs, backend_model.loc_techs_investment_cost, within=po.Reals)

    if 'loc_techs_purchase' in model_data_dict['sets'] and backend_model.mode != 'operate':
        backend_model.purchased = po.Var(backend_model.loc_techs_purchase, within=po.Binary)

    if 'loc_techs_milp' in model_data_dict['sets']:
        if backend_model.mode != 'operate':
            backend_model.units = po.Var(backend_model.loc_techs_milp, within=po.NonNegativeIntegers)
        backend_model.operating_units = po.Var(backend_model.loc_techs_milp, backend_model.timesteps, within=po.NonNegativeIntegers)
        # For any milp tech, we need to update energy_cap, as energy_cap_max and energy_cap_equals
        # are replaced by energy_cap_per_unit
        if backend_model.mode == 'operate':
            for k, v in backend_model.units.items():
                backend_model.energy_cap[k] = v * backend_model.energy_cap_per_unit[k]

    if model_data_dict['attrs'].get('run.ensure_feasibility', False):
        backend_model.unmet_demand = po.Var(backend_model.loc_carriers, backend_model.timesteps, within=po.NonNegativeReals)
        backend_model.bigM = model_data_dict['attrs'].get('run.bigM')



import pytest  # pylint: disable=unused-import
import calliope
import logging
import datetime

from calliope.core.util import dataset

from calliope.core.util.tools import \
    memoize, \
    memoize_instancemethod

from calliope.core.util.logging import log_time


class TestDataset:
    @pytest.fixture()
    def loc_techs(self):
        loc_techs = [
            'region1-3::csp', 'region1::demand_power', 'region1-1::csp',
            'region1-2::csp', 'region2::demand_power', 'region1-1::demand_power'
        ]
        return loc_techs

    def test_get_loc_techs_tech(self, loc_techs):
        loc_techs = dataset.get_loc_techs(loc_techs, tech='csp')
        assert loc_techs == [
            'region1-3::csp', 'region1-1::csp', 'region1-2::csp'
        ]

    def test_get_loc_techs_loc(self, loc_techs):
        loc_techs = dataset.get_loc_techs(loc_techs, loc='region1-1')
        assert loc_techs == [
            'region1-1::csp', 'region1-1::demand_power'
        ]

    def test_get_loc_techs_loc_and_tech(self, loc_techs):
        loc_techs = dataset.get_loc_techs(
            loc_techs, tech='demand_power', loc='region1')
        assert loc_techs == ['region1::demand_power']


class TestMemoization:
    @memoize_instancemethod
    def instance_method(self, a, b):
        return a + b

    def test_memoize_one_arg(self):
        @memoize
        def test(a):
            return a + 1
        assert test(1) == 2
        assert test(1) == 2

    def test_memoize_two_args(self):
        @memoize
        def test(a, b):
            return a + b
        assert test(1, 2) == 3
        assert test(1, 2) == 3

    def test_memoize_instancemethod(self):
        assert self.instance_method(1, 2) == 3
        assert self.instance_method(1, 2) == 3


class TestLogging:
    def test_set_log_level(self):

        # We assign a handler to the Calliope logger on loading calliope
        assert calliope._logger.hasHandlers() is True
        assert len(calliope._logger.handlers) == 1

        for level in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
            calliope.set_log_level(level)
            assert calliope._logger.level == getattr(logging, level)

        # We have a custom level 'SOLVER' at level 19
        calliope.set_log_level('SOLVER')
        assert calliope._logger.level == 19

    def test_timing_log(self):
        timings = {'model_creation': datetime.datetime.now()}

        # TODO: capture logging output and check that comment is in string
        log_time(timings, 'test', comment='test_comment', level='info')
        assert isinstance(timings['test'], datetime.datetime)

        log_time(timings, 'test2', comment=None, level='info')
        assert isinstance(timings['test2'], datetime.datetime)

        # TODO: capture logging output and check that time_since_start is in the string
        log_time(timings, 'test', comment=None, level='info', time_since_start=True)



import os
import tempfile

import pytest  # pylint: disable=unused-import

import calliope
from calliope import exceptions


class TestIO:
    @pytest.fixture(scope='module')
    def model(self):
        model = calliope.examples.national_scale()
        return model

    def test_save_netcdf(self, model):
        bool_attrs = [
            k for k, v in model._model_data.attrs.items()
            if isinstance(v, bool)
        ]

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'model.nc')
            model.to_netcdf(out_path)
            assert os.path.isfile(out_path)

        # Ensure that boolean attrs have not changed
        for k in bool_attrs:
            assert isinstance(model._model_data.attrs[k], bool)

    def test_save_csv_dir_mustnt_exist(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir)
            with pytest.raises(FileExistsError):
                model.to_csv(out_path)

    def test_save_csv(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'out_dir')
            model.to_csv(out_path)
            csv_files = [
                'inputs_resource_eff.csv', 'inputs_reserve_margin.csv', 'inputs_storage_loss.csv',
                'inputs_names.csv', 'inputs_loc_coordinates.csv', 'inputs_lookup_loc_carriers.csv',
                'inputs_lookup_loc_techs_area.csv', 'inputs_energy_con.csv', 'inputs_resource.csv',
                'inputs_storage_cap_max.csv', 'inputs_energy_cap_max_systemwide.csv', 'inputs_cost_om_con.csv',
                'inputs_charge_rate.csv', 'inputs_lifetime.csv', 'inputs_max_demand_timesteps.csv',
                'inputs_parasitic_eff.csv', 'inputs_timestep_resolution.csv', 'inputs_energy_eff.csv',
                'inputs_energy_cap_max.csv', 'inputs_timestep_weights.csv', 'inputs_cost_storage_cap.csv',
                'inputs_cost_resource_cap.csv', 'inputs_cost_depreciation_rate.csv', 'inputs_energy_prod.csv',
                'inputs_cost_resource_area.csv', 'inputs_cost_om_prod.csv', 'inputs_lookup_loc_techs.csv',
                'inputs_cost_energy_cap.csv', 'inputs_colors.csv', 'inputs_resource_unit.csv', 'inputs_inheritance.csv',
                'inputs_energy_ramping.csv', 'inputs_resource_area_max.csv', 'inputs_force_resource.csv'
            ]
            for f in csv_files:
                assert os.path.isfile(os.path.join(out_path, f))

            with open(os.path.join(out_path, 'inputs_energy_cap_max_systemwide.csv'), 'r') as f:
                assert 'demand_power' not in f.read()

    def test_save_csv_no_dropna(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'out_dir')
            model.to_csv(out_path, dropna=False)

            with open(os.path.join(out_path, 'inputs_energy_cap_max_systemwide.csv'), 'r') as f:
                assert 'demand_power' in f.read()

    def test_save_csv_not_optimal(self):
        override_file = os.path.join(
            calliope.examples._PATHS['national_scale'],
            'overrides.yaml'
        )

        # Not checking for content of warnings here, since
        # check_feasibility-related warnings are tested for in
        # test_example_models
        with pytest.warns(exceptions.ModelWarning):
            model = calliope.examples.national_scale(
                override_file=override_file + ':check_feasibility'
            )

        model.run()

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'out_dir')
            with pytest.warns(exceptions.ModelWarning):
                model.to_csv(out_path, dropna=False)

    def test_solve_save_read_netcdf(self, model):
        model.run()

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, 'model.nc')
            model.to_netcdf(out_path)
            assert os.path.isfile(out_path)

            model_from_disk = calliope.read_netcdf(out_path)
            # FIXME test for some data in model_from_disk



"""
Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

util.py
~~~~~~~~

Plotting util functions.

"""

import numpy as np
from IPython import get_ipython


def type_of_script():
    if get_ipython() is not None:
        ipy_str = str(type(get_ipython()))
        if 'zmqshell' in ipy_str:
            return 'jupyter'
        if 'terminal' in ipy_str:
            return 'ipython'
    else:
        return 'terminal'


def get_data_layout(
        get_var_data, get_var_layout,
        relevant_vars, layout, model, dataset,
        subset, sum_dims, squeeze,
        get_var_data_kwargs={},
        get_var_layout_kwargs={}):
    """
    For each dropdown dataset in capacity and timeseries plotting, build the
    plotly data dictionary and the plotly layout dictionary.

    """
    # data_len is used to populate visibility of traces, for dropdown
    data_len = [0]
    data = []
    buttons = []
    # fill trace data and add number of traces per var to 'data_len' for use with
    # visibility. first var in loop has visibility == True by default
    visible = True

    for var in relevant_vars:
        data += get_var_data(
            var, model, dataset, visible, subset, sum_dims, squeeze,
            **get_var_data_kwargs)
        data_len.append(len(data))
        visible = False

    # Initialise all visibility to False for dropdown updates
    total_data_arrays = np.array([False for i in range(data_len[-1])])
    var_num = 0
    for var in relevant_vars:
        # update visibility to True for all traces linked to this variable `var`
        visible_data = total_data_arrays.copy()
        visible_data[data_len[var_num]:data_len[var_num + 1]] = True

        # Get variable-specific layout
        var_layout = get_var_layout(var, dataset, **get_var_layout_kwargs)

        if var_num == 0:
            layout['title'] = var_layout['title']
            if 'barmode' in var_layout:
                layout['barmode'] = var_layout['barmode']

        if len(relevant_vars) > 1:
            var_layout = [{'visible': list(visible_data)}, var_layout]
            buttons.append(dict(label=var, method='update', args=var_layout))

        var_num += 1

    # If there are multiple vars to plot, use dropdowns via 'updatemenus'
    if len(relevant_vars) > 1:
        updatemenus = list([dict(
            active=0, buttons=buttons, type='dropdown',
            xanchor='left', x=0, y=1.13, pad=dict(t=0.05, b=0.05, l=0.05, r=0.05)
        )])
        layout['updatemenus'] = updatemenus
    else:
        layout.update(var_layout)
    return data, layout


def hex_to_rgba(hex_color, opacity):
    """Embed opacity in a colour by converting calliope HEX colours to an RGBA"""
    _NUMERALS = '0123456789abcdefABCDEF'
    _HEXDEC = {v: int(v, 16) for v in (x + y for x in _NUMERALS for y in _NUMERALS)}
    hex_color = hex_color.lstrip('#')
    rgb = [_HEXDEC[hex_color[0:2]], _HEXDEC[hex_color[2:4]], _HEXDEC[hex_color[4:6]]]
    return 'rgba({1}, {2}, {3}, {0})'.format(opacity, *rgb)


def break_name(name):
    """Take a long technology name string and break it across multiple lines"""
    name_breaks = len(name) / 30
    if name_breaks > 1:
        initial_name = name
        break_buffer = 0
        for name_break in range(1, int(name_breaks) + 1):
            # preferably break at a space
            breakpoint = initial_name.rfind(' ', name_break * 30 - 10, name_break * 30)
            # -1 means rfind failed to find any space
            if breakpoint != -1:
                breakpoint += break_buffer
                name = name[:breakpoint].rstrip() + '<br>' + name[breakpoint:].lstrip()
                break_buffer += 4
            else:
                breakpoint = 30 * name_break + break_buffer
                name = name[:breakpoint].rstrip() + '...<br>' + name[breakpoint:].lstrip()
                break_buffer += 7

    return name


def get_clustered_layout(dataset):
    layout = {}
    timestep_cluster = dataset.timestep_cluster.to_pandas()
    clusters = timestep_cluster.groupby(timestep_cluster).groups
    layout['xaxis'] = {}
    layout['xaxis']['type'] = 'category'
    layout['xaxis']['tickvals'] = [
        (2 * (k - min(clusters.keys())) + 1) * len(v) / 2 - 0.5
        for k, v in clusters.items()
    ]
    layout['xaxis']['ticktext'] = [k for k in clusters.keys()]
    layout['xaxis']['title'] = 'Clusters'

    # Make rectangles to fit in the background over every other cluster,
    # to distinguish them
    layout['shapes'] = []
    shape_template = {
        'type': 'rect', 'xref': 'x', 'yref': 'paper', 'y0': 0, 'y1': 1,
        'line': {'width': 0}, 'layer': 'below'
    }

    for cluster in clusters.keys():
        x0 = clusters[cluster][0]
        x1 = clusters[cluster][-1]
        opacity = 0.3 * (cluster % 2)
        day_shape = {'x0': x0, 'x1': x1, 'fillcolor': 'grey', 'opacity': opacity}

        shape_template.update(day_shape)
        layout['shapes'].append(shape_template.copy())

    return layout
